#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "../GPU_Lib.h"
#include "../GPU_ExpEval.h"
#include "../SpeedTest/Utils.h" // p_LoadImage, Save_RGBLinearImageStruct

int main()
{
	if(!Init_OpenGL()) {
		fprintf(stderr, "error: OpenGL initialization failed\n");
		return -1;
	}
	// init openGL

	ImageStruct *p_src0, *p_src1;
	if((p_src0 = p_LoadImage("_lena_256.jpg")) == NULL ||
	   (p_src1 = p_LoadImage("_edgefinder_a.tga")) == NULL)
		return -1;
	// load input images

	ImageStruct *p_result, *p_temp;
	if((p_result = p_Alloc_Image(p_src0->XSize, p_src0->YSize, p_src0->PixelType)) == NULL ||
	   (p_temp = p_Alloc_Image(p_src0->XSize, p_src0->YSize, p_src0->PixelType)) == NULL) {
		fprintf(stderr, "error: error creating ImageStruct for result\n");
		return -1;
	}
	// allocate dest and intermediate images

	float p_set_matrix[] = {
		1, 0, 0, 0,
		1, 0, 0, 0,
		1, 0, 0, 0,
		1, 0, 0, 0
	};
	// matrix for setting all color planes equal

	float p_sum_matrix[] = {
		1, 1, 1, 1,
		1, 1, 1, 1,
		1, 1, 1, 1,
		0, 0, 0, 0
	};
	// matrix for summing up color planes

	Disable_AutoDownloads();
	Disable_AutoUploads();
	// disable downloads and uploads

	Set_DirtyFlag(p_src0, false);
	Set_DirtyFlag(p_src1, false);
	// make sure source images will download (first time they're used only)

	ImageGreyscale(p_temp, p_src0);
	ImageThresh(p_result, p_temp, .5);
	ImageColorMatrix(p_temp, p_result, p_set_matrix, false); // false = not-transpose
	ImageImageHitMiss(p_result, p_temp, p_src1);
	ImageColorMatrix(p_temp, p_result, p_sum_matrix, false); // false = not-transpose
	ImageInvert(p_result, p_temp);
	// calculate the expression

	Download_ImageStruct(p_result);
	// download image

	if(!Save_RGBLinearImageStruct("edges_lena.bmp", p_result))
		fprintf(stderr, "error: error writing output image\n");

	p_temp->Optional.Regular.Free(p_temp);
	// delete temp image

	p_src0->Optional.Regular.Free(p_src0);
	p_src1->Optional.Regular.Free(p_src1);
	p_result->Optional.Regular.Free(p_result);
	// defete other images (in case GPU_Lib allocated them,
	// don't use your own code to free them)

	if(!Shutdown_OpenGL()) {
		fprintf(stderr, "error: OpenGL shutdown failed");
		return -1;
	}

	return 0;
}
