#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "../GPU_Lib.h"
#include "../GPU_ExpEval.h"
#include "../SpeedTest/Utils.h" // p_LoadImage, Save_RGBLinearImageStruct

int main()
{
	if(!Init_OpenGL()) {
		fprintf(stderr, "error: OpenGL initialization failed\n");
		return -1;
	}
	// init openGL

	ImageStruct *p_src;
	if((p_src = p_LoadImage("_lena_256.jpg")) == NULL)
		return -1;
	// load input images

	ImageStruct *p_result, *p_temp;
	if((p_result = p_Alloc_Image(p_src->XSize, p_src->YSize, ImageComplexLinear)) == NULL ||
	   (p_temp = p_Alloc_Image(p_src->XSize, p_src->YSize, ImageRGBLinear)) == NULL) {
		fprintf(stderr, "error: error creating ImageStruct for result\n");
		return -1;
	}
	// allocate dest and intermediate images

	Disable_AutoDownloads();
	Disable_AutoUploads();
	// disable downloads and uploads

	FFTHandle h_handle;
	if((h_handle = h_CreateFFT(p_src->XSize, true)) == InvalidFFTHandle)
	// create forward FFT handle (just one, source width and height are equal)

	ImageFFT(p_result, p_src, h_handle, h_handle);
	// calculate FFT

	DeleteFFT(h_handle);
	// delete FFT

	ImageScale(p_temp, p_result, 1, 0);
	// scale fft and convert to RGB simultaneously

	Download_ImageStruct(p_temp);
	// download image

	if(!Save_RGBLinearImageStruct("fft_lena.bmp", p_temp))
		fprintf(stderr, "error: error writing output image\n");

	p_temp->Optional.Regular.Free(p_temp);
	// delete temp image

	p_src->Optional.Regular.Free(p_src);
	p_result->Optional.Regular.Free(p_result);
	// defete other images (in case GPU_Lib allocated them,
	// don't use your own code to free them)

	if(!Shutdown_OpenGL()) {
		fprintf(stderr, "error: OpenGL shutdown failed");
		return -1;
	}

	return 0;
}
