/*
								+----------------------------------+
								|                                  |
								|   ***  Exp-eval functions  ***   |
								|                                  |
								|   Copyright   -tHE SWINe- 2006  |
								|                                  |
								|         GPU_ExpFuncs.cpp         |
								|                                  |
								+----------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include "GPU_Lib.h"
#include "GPU_ExpEval.h"
#include "GPU_ExpFuncs.h"
#include "GPU_ExpParser.h"

#if defined(_MSC_VER) && !defined(__MWERKS__)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/*
 * here are functions which are used to execute expression tree nodes.
 * most of them directly calls GPU_Lib with node data as parameters.
 */

bool CopyImgPtr(TExpression *p_expr, void **p_params)
{
	return p_expr->n_variable_id != -1 &&
		Set_DirtyFlag(p_expr->t_result.p_image =
		(ImageStruct*)p_params[p_expr->n_variable_id], false); // make sure it uploads
}

bool CopyVarPtr(TExpression *p_expr, void **p_params)
{
	return p_expr->n_variable_id != -1 &&
		(p_expr->t_result.p_void_ptr = p_params[p_expr->n_variable_id]) != NULL;
}

bool CopyVarFloat(TExpression *p_expr, void **p_params)
{
	return p_expr->n_variable_id != -1 &&
		(p_expr->t_result.f_scalar = *(float*)(p_params[p_expr->n_variable_id]));
}

bool CallInverti(TExpression *p_expr, void **p_params)
{
	return ImageInvert(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallAbsi(TExpression *p_expr, void **p_params)
{
	return ImageAbs(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallLogi(TExpression *p_expr, void **p_params)
{
	return ImageLog(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallExpi(TExpression *p_expr, void **p_params)
{
	return ImageExp(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallSqrti(TExpression *p_expr, void **p_params)
{
	return ImageSqrt(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallGreyi(TExpression *p_expr, void **p_params)
{
	return ImageGreyscale(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallBlendiinn(TExpression *p_expr, void **p_params)
{
	return ImageImageBlend(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image,
		p_expr->p_sub_expr[2]->t_result.f_scalar,
		p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallMini(TExpression *p_expr, void **p_params)
{
	unsigned char p_temp[16], *p_result = &p_temp[0];
	if(!ImageGetMinMax(p_expr->p_sub_expr[0]->t_result.p_image, p_result))
		return false;
	TVector4 t_vector;
	switch(p_expr->p_sub_expr[0]->t_result.p_image->PixelType) {
	case Image8:
	case Image8Linear:
		t_vector[0] = t_vector[1] = t_vector[2] = p_result[0] / 255.0f;
		t_vector[3] = 1;
		break;
	case Image16:
	case Image16Linear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((short*)p_result)[0] / 32768.0f;
		t_vector[3] = 1;
		break;
	case Image32:
	case Image32Linear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((int*)p_result)[0] / (float)(1 << 31);
		t_vector[3] = 1;
		break;
	case ImageFloat:
	case ImageFloatLinear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((float*)p_result)[0];
		t_vector[3] = 1;
		break;
	case ImageRGB:
	case ImageRGBLinear:
		{
			unsigned int n_pixel = ((int*)p_result)[0];
			t_vector[0] = ((n_pixel >> 16) & 0xff) / 255.0f;
			t_vector[1] = ((n_pixel >> 8) & 0xff) / 255.0f;
			t_vector[2] = (n_pixel & 0xff) / 255.0f;
			t_vector[3] = (n_pixel >> 24) / 255.0f;
		}
		break;
	case ImageComplex:
	case ImageComplexLinear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((ImageComplexValue*)p_result)[0].Re;
		t_vector[3] = ((ImageComplexValue*)p_result)[0].Im;
		break;
	default:
		return false;
	}
	memcpy(p_expr->t_result.p_vector, &t_vector, sizeof(TVector4));
	return true;
}

bool CallMinii(TExpression *p_expr, void **p_params)
{
	return ImageImageMin(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallMininn(TExpression *p_expr, void **p_params)
{
	return ImageLocalMin(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[1]->t_result.f_scalar,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar);
}

bool CallMaxi(TExpression *p_expr, void **p_params)
{
	unsigned char p_temp[16], *p_result = &p_temp[0];
	if(!ImageGetMinMax(p_expr->p_sub_expr[0]->t_result.p_image, p_result))
		return false;
	TVector4 t_vector;
	switch(p_expr->p_sub_expr[0]->t_result.p_image->PixelType) {
	case Image8:
	case Image8Linear:
		t_vector[0] = t_vector[1] = t_vector[2] = p_result[1] / 255.0f;
		t_vector[3] = 1;
		break;
	case Image16:
	case Image16Linear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((short*)p_result)[1] / 32768.0f;
		t_vector[3] = 1;
		break;
	case Image32:
	case Image32Linear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((int*)p_result)[1] / (float)(1 << 31);
		t_vector[3] = 1;
		break;
	case ImageFloat:
	case ImageFloatLinear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((float*)p_result)[1];
		t_vector[3] = 1;
		break;
	case ImageRGB:
	case ImageRGBLinear:
		{
			unsigned int n_pixel = ((int*)p_result)[1];
			t_vector[0] = ((n_pixel >> 16) & 0xff) / 255.0f;
			t_vector[1] = ((n_pixel >> 8) & 0xff) / 255.0f;
			t_vector[2] = (n_pixel & 0xff) / 255.0f;
			t_vector[3] = (n_pixel >> 24) / 255.0f;
		}
		break;
	case ImageComplex:
	case ImageComplexLinear:
		t_vector[0] = t_vector[1] = t_vector[2] = ((ImageComplexValue*)p_result)[1].Re;
		t_vector[3] = ((ImageComplexValue*)p_result)[1].Im;
		break;
	default:
		return false;
	}
	memcpy(p_expr->t_result.p_vector, &t_vector, sizeof(TVector4));
	return true;
}

bool CallMaxii(TExpression *p_expr, void **p_params)
{
	return ImageImageMax(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallMaxinn(TExpression *p_expr, void **p_params)
{
	return ImageLocalMax(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[1]->t_result.f_scalar,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar);
}

bool CallThreshin(TExpression *p_expr, void **p_params)
{
	return ImageThresh(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.f_scalar);
}

bool CallPowin(TExpression *p_expr, void **p_params)
{
	return ImagePow(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.f_scalar);
}

bool CallMedianin(TExpression *p_expr, void **p_params)
{
	return ImageMedian(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[1]->t_result.f_scalar);
}

bool CallScaleinn(TExpression *p_expr, void **p_params)
{
	return ImageScale(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.f_scalar, p_expr->p_sub_expr[2]->t_result.f_scalar);
}

bool CallMeaninn(TExpression *p_expr, void **p_params)
{
	return ImageMean(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[1]->t_result.f_scalar,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar);
}

bool CallCmatrixim(TExpression *p_expr, void **p_params)
{
	return ImageColorMatrix(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		&(*p_expr->p_sub_expr[1]->t_result.p_matrix)[0][0], false);
}

bool CallDecalii(TExpression *p_expr, void **p_params)
{
	return ImageImageDecal(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallHitMisii(TExpression *p_expr, void **p_params)
{
	return ImageImageHitMiss(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallErodeii(TExpression *p_expr, void **p_params)
{
	return ImageImageErode(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallDilateii(TExpression *p_expr, void **p_params)
{
	return ImageImageDilate(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallOpenii(TExpression *p_expr, void **p_params)
{
	return ImageImageOpen(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallCloseii(TExpression *p_expr, void **p_params)
{
	return ImageImageClose(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallErodeiSquarenn(TExpression *p_expr, void **p_params)
{
	return ImageSquareErode(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Square!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallDilateiSquarenn(TExpression *p_expr, void **p_params)
{
	return ImageSquareDilate(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Square!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallOpeniSquarenn(TExpression *p_expr, void **p_params)
{
	return ImageSquareOpen(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Square!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallCloseiSquarenn(TExpression *p_expr, void **p_params)
{
	return ImageSquareClose(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Square!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallErodeiDiamondnn(TExpression *p_expr, void **p_params)
{
	return ImageDiamondErode(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Diamond!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallDilateiDiamondnn(TExpression *p_expr, void **p_params)
{
	return ImageDiamondDilate(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Diamond!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallOpeniDiamondnn(TExpression *p_expr, void **p_params)
{
	return ImageDiamondOpen(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Diamond!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallCloseiDiamondnn(TExpression *p_expr, void **p_params)
{
	return ImageDiamondClose(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		(int)p_expr->p_sub_expr[2]->t_result.f_scalar, // p_sub_expr[1] = token_Diamond!
		(int)p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallConvolveii(TExpression *p_expr, void **p_params)
{
	return ImageConvolve2D(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallConvolveiii(TExpression *p_expr, void **p_params)
{
	return ImageConvolveSeparable(p_expr->t_result.p_image,
		p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image, p_expr->p_sub_expr[2]->t_result.p_image);
}

bool CallResizei(TExpression *p_expr, void **p_params)
{
	return ImageResize(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallResizeinn(TExpression *p_expr, void **p_params)
{
	int n_width = (int)p_expr->p_sub_expr[1]->t_result.f_scalar;
	int n_height = (int)p_expr->p_sub_expr[2]->t_result.f_scalar;
	if(p_expr->t_result.p_image == NULL || p_expr->t_result.p_image->XSize != n_width ||
	   p_expr->t_result.p_image->YSize != n_height) {
		if(p_expr->t_result.p_image != NULL)
			p_expr->t_result.p_image->Optional.Regular.Free(p_expr->t_result.p_image);
		if((p_expr->t_result.p_image = p_Alloc_Image(n_width, n_height,
		   p_expr->p_sub_expr[0]->t_result.p_image->PixelType)) == NULL)
			return false;
	}
	return ImageResize(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image);
}

bool CallLerpi8n(TExpression *p_expr, void **p_params)
{
	return ImageTransform(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.f_scalar, p_expr->p_sub_expr[2]->t_result.f_scalar,
		p_expr->p_sub_expr[3]->t_result.f_scalar, p_expr->p_sub_expr[4]->t_result.f_scalar,
		p_expr->p_sub_expr[5]->t_result.f_scalar, p_expr->p_sub_expr[6]->t_result.f_scalar,
		p_expr->p_sub_expr[7]->t_result.f_scalar, p_expr->p_sub_expr[8]->t_result.f_scalar);
}

bool CallLerpinn8n(TExpression *p_expr, void **p_params)
{
	int n_width = (int)p_expr->p_sub_expr[1]->t_result.f_scalar;
	int n_height = (int)p_expr->p_sub_expr[2]->t_result.f_scalar;
	if(p_expr->t_result.p_image == NULL || p_expr->t_result.p_image->XSize != n_width ||
	   p_expr->t_result.p_image->YSize != n_height) {
		if(p_expr->t_result.p_image != NULL)
			p_expr->t_result.p_image->Optional.Regular.Free(p_expr->t_result.p_image);
		if((p_expr->t_result.p_image = p_Alloc_Image(n_width, n_height,
		   p_expr->p_sub_expr[0]->t_result.p_image->PixelType)) == NULL)
			return false;
	}
	return ImageTransform(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[3]->t_result.f_scalar, p_expr->p_sub_expr[4]->t_result.f_scalar,
		p_expr->p_sub_expr[5]->t_result.f_scalar, p_expr->p_sub_expr[6]->t_result.f_scalar,
		p_expr->p_sub_expr[7]->t_result.f_scalar, p_expr->p_sub_expr[8]->t_result.f_scalar,
		p_expr->p_sub_expr[9]->t_result.f_scalar, p_expr->p_sub_expr[10]->t_result.f_scalar);
}

bool CallRotateinbn(TExpression *p_expr, void **p_params)
{
	return ImageRotate(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.f_scalar, p_expr->p_sub_expr[2]->n_type == token_True,
		p_expr->p_sub_expr[3]->t_result.f_scalar);
}

bool CallRotateinnnbn(TExpression *p_expr, void **p_params)
{
	int n_width = (int)p_expr->p_sub_expr[1]->t_result.f_scalar;
	int n_height = (int)p_expr->p_sub_expr[2]->t_result.f_scalar;
	if(p_expr->t_result.p_image == NULL || p_expr->t_result.p_image->XSize != n_width ||
	   p_expr->t_result.p_image->YSize != n_height) {
		if(p_expr->t_result.p_image != NULL)
			p_expr->t_result.p_image->Optional.Regular.Free(p_expr->t_result.p_image);
		if((p_expr->t_result.p_image = p_Alloc_Image(n_width, n_height,
		   p_expr->p_sub_expr[0]->t_result.p_image->PixelType)) == NULL)
			return false;
	}
	return ImageRotate(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[3]->t_result.f_scalar, p_expr->p_sub_expr[4]->n_type == token_True,
		p_expr->p_sub_expr[5]->t_result.f_scalar);
}

bool CallAddin(TExpression *p_expr, void **p_params)
{
	return ImageScale(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image, 1.0,
		p_expr->p_sub_expr[1]->t_result.f_scalar);
}

bool CallAddiv(TExpression *p_expr, void **p_params)
{
	float p_mat4x4[16] = {
		1, 0, 0, (*p_expr->p_sub_expr[1]->t_result.p_vector)[0],
		0, 1, 0, (*p_expr->p_sub_expr[1]->t_result.p_vector)[1],
		0, 0, 1, (*p_expr->p_sub_expr[1]->t_result.p_vector)[2],
		0, 0, 0, 1 + (*p_expr->p_sub_expr[1]->t_result.p_vector)[3]
	};
	return ImageColorMatrix(p_expr->t_result.p_image,
		p_expr->p_sub_expr[0]->t_result.p_image, p_mat4x4, false);
}

bool CallAddii(TExpression *p_expr, void **p_params)
{
	return ImageImageAdd(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallSubin(TExpression *p_expr, void **p_params)
{
	return ImageScale(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image, 1.0,
		-p_expr->p_sub_expr[1]->t_result.f_scalar);
}

bool CallSubiv(TExpression *p_expr, void **p_params)
{
	float p_mat4x4[16] = {
		1, 0, 0, -(*p_expr->p_sub_expr[1]->t_result.p_vector)[0],
		0, 1, 0, -(*p_expr->p_sub_expr[1]->t_result.p_vector)[1],
		0, 0, 1, -(*p_expr->p_sub_expr[1]->t_result.p_vector)[2],
		0, 0, 0, 1 - (*p_expr->p_sub_expr[1]->t_result.p_vector)[3]
	};
	return ImageColorMatrix(p_expr->t_result.p_image,
		p_expr->p_sub_expr[0]->t_result.p_image, p_mat4x4, false);
}

bool CallSubii(TExpression *p_expr, void **p_params)
{
	return ImageImageSub(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallMulin(TExpression *p_expr, void **p_params)
{
	return ImageScale(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.f_scalar, 0);
}

bool CallMuliv(TExpression *p_expr, void **p_params)
{
	float p_mat4x4[16] = {
		(*p_expr->p_sub_expr[1]->t_result.p_vector)[0], 0, 0, 0,
		0, (*p_expr->p_sub_expr[1]->t_result.p_vector)[1], 0, 0,
		0, 0, (*p_expr->p_sub_expr[1]->t_result.p_vector)[2], 0,
		0, 0, 0, (*p_expr->p_sub_expr[1]->t_result.p_vector)[3]
	};
	return ImageColorMatrix(p_expr->t_result.p_image,
		p_expr->p_sub_expr[0]->t_result.p_image, p_mat4x4, false);
}

bool CallMulii(TExpression *p_expr, void **p_params)
{
	return ImageImageMult(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

bool CallDivin(TExpression *p_expr, void **p_params)
{
	return ImageScale(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		1.0f / p_expr->p_sub_expr[1]->t_result.f_scalar, 0);
}

bool CallDiviv(TExpression *p_expr, void **p_params)
{
	float p_mat4x4[16] = {
		1.0f / (*p_expr->p_sub_expr[1]->t_result.p_vector)[0], 0, 0, 0,
		0, 1.0f / (*p_expr->p_sub_expr[1]->t_result.p_vector)[1], 0, 0,
		0, 0, 1.0f / (*p_expr->p_sub_expr[1]->t_result.p_vector)[2], 0,
		0, 0, 0, 1.0f / (*p_expr->p_sub_expr[1]->t_result.p_vector)[3]
	};
	return ImageColorMatrix(p_expr->t_result.p_image,
		p_expr->p_sub_expr[0]->t_result.p_image, p_mat4x4, false);
}

bool CallDivii(TExpression *p_expr, void **p_params)
{
	return ImageImageDiv(p_expr->t_result.p_image, p_expr->p_sub_expr[0]->t_result.p_image,
		p_expr->p_sub_expr[1]->t_result.p_image);
}

/*
 *		-end-of-file-
 */
