/*
								+----------------------------------+
								|                                  |
								|   ***  Exp-eval functions  ***   |
								|                                  |
								|   Copyright   -tHE SWINe- 2006  |
								|                                  |
								|         GPU_ExpFuncs.cpp         |
								|                                  |
								+----------------------------------+
*/

#ifndef __EXPRESSION_EXEC_FUNCTIONS_INCLUDED
#define __EXPRESSION_EXEC_FUNCTIONS_INCLUDED

/*
 * here are functions which are used to execute expression tree nodes.
 * most of them directly calls GPU_Lib with node data as parameters.
 */

/**
 *	struct TFunctionBinding
 *		- binding of execution function to function name and parameter types / count
 */
typedef struct {
	int n_function_token;
	int n_return_type;
	int n_param_num;
	int p_param_type[16];
	bool (*p_exec_func)(TExpression *p_expr, void **p_params);
} TFunctionBinding;

bool CopyImgPtr(TExpression *p_expr, void **p_params);
bool CopyVarPtr(TExpression *p_expr, void **p_params);
bool CopyVarFloat(TExpression *p_expr, void **p_params);
bool CallInverti(TExpression *p_expr, void **p_params);
bool CallAbsi(TExpression *p_expr, void **p_params);
bool CallLogi(TExpression *p_expr, void **p_params);
bool CallExpi(TExpression *p_expr, void **p_params);
bool CallSqrti(TExpression *p_expr, void **p_params);
bool CallGreyi(TExpression *p_expr, void **p_params);
bool CallBlendiinn(TExpression *p_expr, void **p_params);
bool CallMini(TExpression *p_expr, void **p_params);
bool CallMinii(TExpression *p_expr, void **p_params);
bool CallMininn(TExpression *p_expr, void **p_params);
bool CallMaxi(TExpression *p_expr, void **p_params);
bool CallMaxii(TExpression *p_expr, void **p_params);
bool CallMaxinn(TExpression *p_expr, void **p_params);
bool CallThreshin(TExpression *p_expr, void **p_params);
bool CallPowin(TExpression *p_expr, void **p_params);
bool CallMedianin(TExpression *p_expr, void **p_params);
bool CallScaleinn(TExpression *p_expr, void **p_params);
bool CallMeaninn(TExpression *p_expr, void **p_params);
bool CallCmatrixim(TExpression *p_expr, void **p_params);
bool CallDecalii(TExpression *p_expr, void **p_params);
bool CallHitMisii(TExpression *p_expr, void **p_params);
bool CallErodeii(TExpression *p_expr, void **p_params);
bool CallDilateii(TExpression *p_expr, void **p_params);
bool CallOpenii(TExpression *p_expr, void **p_params);
bool CallCloseii(TExpression *p_expr, void **p_params);
bool CallErodeiSquarenn(TExpression *p_expr, void **p_params);
bool CallDilateiSquarenn(TExpression *p_expr, void **p_params);
bool CallOpeniSquarenn(TExpression *p_expr, void **p_params);
bool CallCloseiSquarenn(TExpression *p_expr, void **p_params);
bool CallErodeiDiamondnn(TExpression *p_expr, void **p_params);
bool CallDilateiDiamondnn(TExpression *p_expr, void **p_params);
bool CallOpeniDiamondnn(TExpression *p_expr, void **p_params);
bool CallCloseiDiamondnn(TExpression *p_expr, void **p_params);
bool CallConvolveii(TExpression *p_expr, void **p_params);
bool CallConvolveiii(TExpression *p_expr, void **p_params);
bool CallResizei(TExpression *p_expr, void **p_params);
bool CallResizeinn(TExpression *p_expr, void **p_params);
bool CallLerpi8n(TExpression *p_expr, void **p_params);
bool CallLerpinn8n(TExpression *p_expr, void **p_params);
bool CallRotateinbn(TExpression *p_expr, void **p_params);
bool CallRotateinnnbn(TExpression *p_expr, void **p_params);
bool CallAddin(TExpression *p_expr, void **p_params);
bool CallAddiv(TExpression *p_expr, void **p_params);
bool CallAddii(TExpression *p_expr, void **p_params);
bool CallSubin(TExpression *p_expr, void **p_params);
bool CallSubiv(TExpression *p_expr, void **p_params);
bool CallSubii(TExpression *p_expr, void **p_params);
bool CallMulin(TExpression *p_expr, void **p_params);
bool CallMuliv(TExpression *p_expr, void **p_params);
bool CallMulii(TExpression *p_expr, void **p_params);
bool CallDivin(TExpression *p_expr, void **p_params);
bool CallDiviv(TExpression *p_expr, void **p_params);
bool CallDivii(TExpression *p_expr, void **p_params);

#endif //__EXPRESSION_EXEC_FUNCTIONS_INCLUDED
