/*
								+----------------------------------+
								|                                  |
								| ***  GPU image-proc library  *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|           GPU_Lib.cpp            |
								|                                  |
								+----------------------------------+
*/

#if defined(_MSC_VER) && !defined(__MWERKS__)
#pragma warning(disable:4786)
#endif
// disable warning C4786 "identifier was truncated to '255' characters in the debug information"

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <algorithm>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLDrv.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/VertexBufferObject.h"
#include "../berLame_src/Vector.h"
#include "GPU_Lib.h"
#include "TempTex_Manager.h"
#include "Shader_Manager.h"
#include "RenderBuffer_Manager.h"
#include "GPU_Lib_Private.h"
#include "OpenGL20Logo.h"
#include "ShadersSrc.h"
#include "ImageTransfer.h"
#include "ImageBind.h"
#include "ImageProc.h"
#include "RefRaster.h"

int CGLNamedTexture_2D::m_n_id_space = 0; // id counter for textures
// Texture2D_manager

int CGLNamedRenderBuffer_FBO::m_n_id_space = 0; // id counter for render-buffers
// RenderBuffer_Manager

int CGLNamedShaderObject::m_n_id_space = 0; // id counter for shaders
std::vector<CShaderFactory::TShaderSource> CShaderFactory::m_shader_source_list; // list of shader sources
// Shader_Manager

static CGLState *p_state = 0;
// opengl state cache

static bool b_pixel_buffer_supported = false;
static bool b_fence_supported = false;
static bool b_fbo_supported = false;
static bool b_imaging_supported = false;
static bool b_gl_images_supported = false;
static bool b_shaders_supported = false;
// opengl capabilities

static std::vector<TImageStructInfo> imagestruct_list;
// list of textures, bound to imagestructs

static bool b_auto_uploads = true;
static bool b_auto_downloads = true;
// automatic uploads / downloads (not static for ImageBind.cpp)

CGLRenderBuffer_Manager render_buffer_manager;
// render buffer manager object

CGLShader_Manager shader_manager;
// shader manager object

CGLTexture_2D_Manager temp_texture_manager;
// temporary texture manager object

CGLPixelBufferObject *p_tmp_upload_buffer = 0;
CGLPixelBufferObject *p_tmp_download_buffer = 0;
CGLPixelBufferObject *p_tmp_download_buffer2 = 0;
// temporal buffer for async image transfers

static const char *p_s_OpenGL_window_name = (GPU_LIB_OPENGL_WINDOW_NAME);
static const char *p_s_OpenGL_window_classname = "my_gl_win_class";
// window name and wndclass

/*
 *								--- managers access for other modules ---
 */

/**
 *	CGLNamedTexture_2D *p_GetTemporaryTexture2D(TTextureFormat t_format)
 *		- returns temporary 2D texture with format t_format or 0 in case error occured
 */
CGLNamedTexture_2D *p_GetTemporaryTexture2D(TTextureFormat t_format)
{
	int n_texture_id;
	if((n_texture_id = temp_texture_manager.n_UnusedObject_Id(t_format, -1)) == -1 &&
	   (!temp_texture_manager.LoadObject(t_format, CTextureFactory(p_state)) ||
	   (n_texture_id = temp_texture_manager.n_UnusedObject_Id(t_format, -1)) == -1))
		return 0;
	CGLNamedTexture_2D *p_texture = &temp_texture_manager.r_Object(n_texture_id);
	temp_texture_manager.ReferenceObject(n_texture_id, false);
	return p_texture;
}

/**
 *	void ReturnTemporaryTexture(int n_texture_id)
 *		- returns texture n_tex_id back to circulation
 */
void ReturnTemporaryTexture(int n_texture_id)
{
	temp_texture_manager.ReturnObject(n_texture_id, false);
}

/**
 *	CGLNamedShaderObject *p_GetShader(const char *p_s_shader_name)
 *		- gets compiled shader with name p_s_shader_name or 0 in case error occured
 */
CGLNamedShaderObject *p_GetShader(const char *p_s_shader_name)
{
	int n_shader_id;
	if((n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1 &&
	   (!shader_manager.LoadObject(p_s_shader_name, CShaderFactory(p_state)) ||
	   (n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1))
		return 0;
	return &shader_manager.r_Object(n_shader_id);
}

/**
 *	CGLNamedRenderBuffer_FBO *p_GetRenderBuffer(const CGLTexture_2D *p_texture)
 *		- returns temporary render suitable buffer for rendering into 2D
 *		  texture with format of p_texture
 *		- returns 0 in case error occured
 */
CGLNamedRenderBuffer_FBO *p_GetRenderBuffer(const CGLTexture_2D *p_texture)
{
	int n_fbo_id = render_buffer_manager.n_UnusedObject_Id(TTextureFormat(p_texture), -1);
	if(n_fbo_id == -1) {
		if(!render_buffer_manager.LoadObject(TTextureFormat(p_texture),
		   CRenderBufferFactory()))
			return 0;
		n_fbo_id = render_buffer_manager.n_UnusedObject_Id(TTextureFormat(p_texture), -1);
	}
	_ASSERTE(n_fbo_id != -1);
	// find / create appropriate render-buffer

	CGLNamedRenderBuffer_FBO *p_render_buffer = &render_buffer_manager.r_Object(n_fbo_id);
	render_buffer_manager.ReferenceObject(n_fbo_id, false);
	// get our dreamy render buffer

	return p_render_buffer;
}

/**
 *	void ReturnRenderBuffer(int n_fbo_id)
 *		- returns render buffer n_fbo_id back to circulation
 */
void ReturnRenderBuffer(int n_fbo_id)
{
	render_buffer_manager.ReturnObject(n_fbo_id, false);
}

/**
 *	TImageStructInfo *p_FindImageInfo(const ImageStruct *p_image)
 *		- finds existing image info for imagestruct p_image
 *		- returns 0 in case p_image doesn't have image info
 */
TImageStructInfo *p_FindImageInfo(const ImageStruct *p_image)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) != imagestruct_list.end())
		return &*p_image_info;
	return 0;
}

/*
 *								--- ~managers access for other modules ---
 */

/*
 *								--- win32 OpenGL ---
 */

#ifdef WIN32

static HWND h_wnd;
static CGLDriver *p_gl_driver = 0;

/**
 *	bool Create_Window(const char *p_s_name, int n_width, int n_height)
 *		- creates window with given name and size
 *		- returns true on success, false on failure
 */
bool Create_Window(const char *p_s_name, int n_width, int n_height)
{
	WNDCLASSEX t_wnd_class;
	t_wnd_class.cbSize = sizeof(WNDCLASSEX);
	t_wnd_class.style = CS_HREDRAW | CS_VREDRAW;
	t_wnd_class.lpfnWndProc = DefWindowProc;
	t_wnd_class.cbClsExtra = 0;
	t_wnd_class.cbWndExtra = 0;
	t_wnd_class.hInstance = GetModuleHandle(NULL);
	t_wnd_class.hIcon = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	t_wnd_class.hCursor = LoadCursor(GetModuleHandle(NULL), IDC_ARROW);
	t_wnd_class.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	t_wnd_class.lpszMenuName = 0;
	t_wnd_class.lpszClassName = p_s_OpenGL_window_classname;
	t_wnd_class.hIconSm = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	if(!RegisterClassEx(&t_wnd_class))
		return false;
	// create window class

	RECT t_rect;
	t_rect.left = t_rect.top = 0;
	t_rect.right = n_width;
	t_rect.bottom = n_height;
	if(!AdjustWindowRect(&t_rect, WS_BORDER | WS_CAPTION, false))
		return false;
	// calculate window size to contain client area with desired resolution

	if(!(h_wnd = CreateWindow(p_s_OpenGL_window_classname, p_s_name,
	   WS_BORDER | WS_CAPTION, CW_USEDEFAULT, CW_USEDEFAULT,
	   t_rect.right - t_rect.left, t_rect.bottom - t_rect.top,
	   NULL, NULL, GetModuleHandle(NULL), NULL)))
		return false;
	// create window

#ifdef GPU_LIB_WIN32_SHOW_WINDOW
	ShowWindow(h_wnd, SW_SHOW);
#else //GPU_LIB_WIN32_SHOW_WINDOW
	ShowWindow(h_wnd, SW_HIDE);
#endif //GPU_LIB_WIN32_SHOW_WINDOW
	UpdateWindow(h_wnd);
	// show & update window

	return true;
}

#endif //WIN32

/*
 *								--- ~win32 OpenGL ---
 */

/*
 *								--- OpenGL init code ---
 */

/**
 *	bool Init_OpenGL()
 *		- initializes OpenGL. under linux uses GLUT, under windows uses custom driver
 *		- in case running under windows, there's option to display OpenGL window,
 *		  containing "OpenGL 2.0" logo (enabled using GPU_LIB_WIN32_SHOW_WINDOW #define)
 *		- returns true on success (or multiple initialization), false on failure
 */
bool Init_OpenGL()
{
	if(p_state)
		return true;

	if(!Register_Shaders())
		return false;
	// register shaders to CShaderFactory

#ifdef WIN32
	if(!Create_Window(p_s_OpenGL_window_name, 150, 57))
		return false;
	// create window

	if(!(p_gl_driver = new CGLDriver()))
		return false;
	if(!p_gl_driver->Init(h_wnd, 150, 57, 24, 0)) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// init open-gl

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext()))) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// create state guard
#else //WIN32
	// t_odo -  linux opengl (init glut)

	int n_arg_num = 1;
	char *p_s_arg = "no-arg";
	glutInit(&n_arg_num, &p_s_arg);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
    glutInitWindowSize(150, 57);
    glutCreateWindow(p_s_OpenGL_window_name);

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext())))
		return false;
	// create state guard
#endif //WIN32

	b_fbo_supported = CGLRenderBuffer_FBO::b_Supported(); // gets function pointers as well
	CGLExtensionHandler::n_GetGL12FuncPointers();
	CGLExtensionHandler::n_GetGL13FuncPointers();
	CGLExtensionHandler::n_GetMultitextureFuncPointers(); // opengl1.3 as well
	CGLExtensionHandler::n_GetGL14FuncPointers();
	CGLExtensionHandler::n_GetGL15FuncPointers();
	CGLExtensionHandler::n_GetGL20FuncPointers();
	b_pixel_buffer_supported = CGLPixelBufferObject::b_Supported(); // gets function pointers as well
	b_gl_images_supported = CGLPixelBufferObject::b_Supported();
	b_fence_supported = CGLExtensionHandler::b_SupportedExtension("GL_NV_fence") &&
		!CGLExtensionHandler::n_GetFenceNVFuncPointers();
	b_imaging_supported = CGLExtensionHandler::b_SupportedExtension("GL_ARB_imaging");
	b_shaders_supported = CGL_ARB_shader::b_Supported() || CGL_Core_shader::b_Supported();
	// get function pointers

	unsigned __int32 *p_gl_logo;
	if((p_gl_logo = COpenGL20Logo::p_DecompressImage())) {
		/*glDrawPixels(n_opengl_logo_2_0_sm_jpg_width, n_opengl_logo_2_0_sm_jpg_height,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);*/ // upside-down
		CGLTexture_2D opengl_tex(p_state, COpenGL20Logo::n_Width(),
			COpenGL20Logo::n_Height(), GL_RGBA8, false, 0, GL_TEXTURE_2D,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);
		delete[] p_gl_logo;

		p_state->DisableDepthTest();
		p_state->DisableCullFace();
		p_state->Color3f(1, 1, 1);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		opengl_tex.Bind_Enable(p_state);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 1);
		glVertex2f(-1, -1);
		glTexCoord2f(0, 0);
		glVertex2f(-1,  1);
		glTexCoord2f(1, 0);
		glVertex2f( 1,  1);
		glTexCoord2f(1, 1);
		glVertex2f( 1, -1);
		glEnd();

		//p_state->BindTexture2D(0); // destroying texture with logo

#ifdef WIN32
		p_gl_driver->Blit();
#else //WIN32
		glutSwapBuffers();
#endif //WIN32
	}
	// load OpenGL logo

	return true;
}

/**
 *	bool Free_OpenGL_Objects()
 *		- de-allocates all OpenGL objects. use only in case other than built-in OpenGL
 *		  initialization code was used and therefore it's not possible to use Shutdown_OpenGL()
 *		  for shutdown
 *		- returns true on success, false in case OpenGL was not running
 */
bool Free_OpenGL_Objects()
{
	if(!p_state)
		return false;

	std::for_each(imagestruct_list.begin(),
		imagestruct_list.end(), CDeleteImageStructInfo(p_state));
	imagestruct_list.clear();
	// delete associated images

	render_buffer_manager.Clear();
	// delete all render-buffers

	shader_manager.Clear();
	// delete all shaders

	temp_texture_manager.Clear();
	// delete all temporal textures

	if(p_tmp_upload_buffer) {
		delete p_tmp_upload_buffer;
		p_tmp_upload_buffer = 0;
	}
	// delete upload buffer

	if(p_tmp_download_buffer) {
		delete p_tmp_download_buffer;
		p_tmp_download_buffer = 0;
	}
	if(p_tmp_download_buffer2) {
		delete p_tmp_download_buffer2;
		p_tmp_download_buffer2 = 0;
	}
	// delete download buffers

	return true;
}

/**
 *	bool Shutdown_OpenGL();
 *		- shuts down OpenGL, previously initialized by Init_OpenGL(), calls Free_OpenGL_Objects()
 *		  to delete all kinds of temporary objects library uses
 *		- returns true on success, false on failure
 *		- note for linux - there's no function to shut down GLUT, it's expected
 *		  to call this just before exitting as it has really no effect on OpenGL, it frees
 *		  OpenGL objects only
 */
bool Shutdown_OpenGL()
{
	if(!p_state)
		return true;

	bool b_result = Free_OpenGL_Objects();

	delete p_state;
	p_state = 0;

#ifdef WIN32
	b_result &= p_gl_driver->Shutdown();
	delete p_gl_driver;
	// shutdown open-gl

	UnregisterClass(p_s_OpenGL_window_classname, GetModuleHandle(NULL));
	DestroyWindow(h_wnd);
	// destroy window
#else
	// t_odo - linux opengl (shutdown glut)
	// no function to shutdown glut :(
#endif

	return b_result;
}

/*
 *								--- ~OpenGL init code ---
 */

/*
 *								--- Managers controll functions ---
 */

/**
 *	unsigned int n_GPU_ImageStruct_Num()
 *		- returns number of ImageStruct copies, residing on GPU
 */
unsigned int n_GPU_ImageStruct_Num()
{
	return imagestruct_list.size();
}

/**
 *	unsigned int n_Shader_Num()
 *		- returns number of compiled shaders
 */
unsigned int n_Shader_Num()
{
	return render_buffer_manager.n_Object_Num();
}

/**
 *	unsigned int n_RenderBuffer_Num()
 *		- returns number of render-buffers
 */
unsigned int n_RenderBuffer_Num()
{
	return shader_manager.n_Object_Num();
}

/**
 *	unsigned int n_TempTexture_Num()
 *		- returns number of temporary textures
 */
unsigned int n_TempTexture_Num()
{
	return temp_texture_manager.n_Object_Num();
}

/**
 *	unsigned int n_TempTexture_Size()
 *		- returns amount of memory taken by temporary textures
 *		- note there are no functions for querying size of shaders nor render-buffers
 *		  because they take up virtually no memory (render-buffers are just contexts,
 *		  not real buffers)
 */
unsigned int n_TempTexture_Size()
{
	return temp_texture_manager.n_Cost_Sum();
}

/**
 *	void Purge_TempTextures_Size(unsigned int n_size)
 *		- free at least n_size bytes of space from temporary textures
 *		- this may come in handy if there is far too much temporary textures
 *		  and processing is slow due to GPU memory swapping
 *		- note temporary textures accumulate with different image formats and sizes
 *		  it was necessary to work with
 */
void Purge_TempTextures_Size(unsigned int n_size)
{
	temp_texture_manager.n_Purge(n_size);
}

/*
 *								--- ~Managers controll functions ---
 */

/*
 *								--- OpenGL capability query functions ---
 */

/**
 *	bool b_NPOTTexturesSupported()
 *		- returns true in case non-power-of-two textures are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_ARB_texture_non_power_of_two (promoted to core in OpenGL 2.0)
 */
bool b_NPOTTexturesSupported()
{
	return CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_non_power_of_two");
}

/**
 *	bool b_FloatTexturesSupported()
 *		- returns true in case float texture pixel-format is supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_ARB_texture_float
 */
bool b_FloatTexturesSupported()
{
	return CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_float");
}

/**
 *	bool b_FramebufferObjectSupported()
 *		- returns true in case frame-buffer objects (FBO's) are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_EXT_framebuffer_object
 *		- in case FBO's aren't supported, the library is forced to work in "software" mode
 */
bool b_FramebufferObjectSupported()
{
	return b_fbo_supported;
}

/**
 *	bool b_ProgrammableShadingSupported()
 *		- returns true in case OpenGL shading language is supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_ARB_shading_language_100, which is core in OpenGL 2.0
 *		- in case shaders aren't supported, the library is forced to work in "software" mode
 */
bool b_ProgrammableShadingSupported()
{
	return b_shaders_supported;
}

/**
 *	bool b_ImagingSubsetSupported()
 *		- returns true in case OpenGL's imaging subset and pixel-buffer objects
 *		  are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension names are GL_ARB_vertex_buffer_object (promoted to core feature in
 *		  OpenGL 1.5), GL_ARB_pixel_buffer_object (core feature in OpenGL 2.1) and
 *		  GL_ARB_imaging (core feature in OpenGL 1.2)
 *		- GPU accelerated histogram and minmax relies on those and can't work without them
 */
bool b_ImagingSubsetSupported()
{
	return b_imaging_supported && b_pixel_buffer_supported && b_fbo_supported;
}

/**
 *	bool b_AsyncTransfersSupported()
 *		- returns true in case asynchronous data transfers are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension names are GL_ARB_vertex_buffer_object (promoted to core feature in
 *		  OpenGL 1.5) and GL_ARB_pixel_buffer_object (core feature in OpenGL 2.1)
 */
bool b_AsyncTransfersSupported()
{
	return b_pixel_buffer_supported;
}

/**
 *	bool b_NVFenceSupported()
 *		- returns true in case fence mechanism is supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- fence mechanism is useful for determining wheter certain OpenGL operations had
 *		  completed, here for querying asynchronous image transfers completeness
 *		- note extension name is GL_NV_fence (not supported by ATI yet)
 *		- without this extension, async transfers lose their sense
 */
bool b_NVFenceSupported()
{
	return b_fence_supported;
}

/**
 *	int n_MaxTextureSize()
 *		- returns maximal 2D texture dimensions handled by current OpenGL implementation
 *		- return value is only valid in case OpenGL was successfuly initialized
 */
int n_MaxTextureSize()
{
	GLint n_size;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &n_size);
	return n_size;
}

/*
 *								--- ~OpenGL capability query functions ---
 */

/*
 *								--- image related functions ---
 */

/**
 *	bool b_FastTransfer(const ImageStruct *p_image)
 *		- determines whether p_image could be transfered using a single block copy command
 *		  or wheter it needs slower per-scanline copy loop
 *		- returns true in case p_image can be transfered the fast path, otherwise false
 *		- note all images with interleaved pixels and/or scanlines will have to be transferd
 *		  using the slow path
 */
bool b_FastTransfer(const ImageStruct *p_image)
{
	int n_pixel_size = n_PixelSize(p_image->PixelType);

	if(abs(p_image->XOffset) != n_pixel_size)
		return false;
	// pixels are interleaved

	if(abs(p_image->YOffset) % n_pixel_size)
	    return false;
	// scanlines are interleaved in bad fashion

	return true;
}

/**
 *	bool b_HasTexture(const ImageStruct *p_image)
 *		- returns true in case p_image has got associated texture on GPU
 *		- always return false in case OpenGL is not running or was not initialized properly
 *		- note in case image doesn't have texture associated, it shall get one in first
 *		  image processing function where it's subjected either as a parameter or destination
 */
bool b_HasTexture(const ImageStruct *p_image)
{
	return std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
		CFindImageStructInfo(p_image)) != imagestruct_list.end();
}

/**
 *	TImageStructInfo *p_CreateImageInfo(ImageStruct *p_image)
 *		- creates a new image info for imagestruct p_image
 *		- note it doesn't check for duplicates
 *		- returns 0 on failure
 */
TImageStructInfo *p_CreateImageInfo(ImageStruct *p_image)
{
	_ASSERTE(std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image)) == imagestruct_list.end());
	// it doesn't already have one, does it?

	imagestruct_list.reserve(imagestruct_list.size() + 1);
	if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
		return 0;
	// make place in the list

	TImageStructInfo t_temp;
	t_temp.b_current_on_server = true;
	t_temp.b_pixel_buffer_object = false;
	t_temp.p_image_struct = p_image;
	t_temp.p_pbo = 0;
	if(!(t_temp.p_texture = p_Create_ImageStruct_Texture(p_state, p_image)))
		return 0;
	// create imagestruct_list.end()

	imagestruct_list.push_back(t_temp);
	// put it into list

	return &*(imagestruct_list.end() - 1);
}

/**
 *	bool Upload_ImageStruct(ImageStruct *p_image)
 *		- forces upload of p_image from system memory to GPU memory
 *		- returns true on success, false on failure
 */
bool Upload_ImageStruct(ImageStruct *p_image) // nahraje obrazek do graficke karty
{
	TImageStructInfo *p_image_info;
	if(!(p_image_info = p_FindImageInfo(p_image))) {
		return p_CreateImageInfo(p_image) != 0;
		// create new associated texture, containing the image
	}
	// make sure there is associated texture

	p_image_info->p_texture->Bind_Enable(p_state);
	if(p_image_info->b_pixel_buffer_object) {
		int n_format, n_data_type, n_internal_format;
		if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
			return false;

		_ASSERTE(p_image_info->p_pbo);
		p_image_info->p_pbo->UnMap(p_state);
		/**/p_state->BindPixelPackBuffer(0);
		// unmap buffer

		p_image_info->p_pbo->BindAs(p_state, GL_PIXEL_UNPACK_BUFFER_ARB);

		glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
		glPixelStorei(GL_UNPACK_ALIGNMENT, n_GL_PixelSize(n_internal_format));
		// set proper transfer mode

		glTexSubImage2D(GL_TEXTURE_2D,
					0, // LOD pyramid level
					0, // x-offset
					0, // y-offset
					p_image_info->p_texture->n_Width(), // width
					p_image_info->p_texture->n_Height(), // height
					n_format,
					n_data_type,
					p_image_info->p_pbo->p_OffsetPointer(0));
		// tell OpenGL to source texture from PBO

		p_state->BindPixelUnPackBuffer(0);
		
		p_image_info->p_pbo->Bind(p_state); // must bind to force primary target
		if(!(p_image->Raster = (unsigned char*)p_image_info->p_pbo->p_Map(p_state,
		   GL_READ_WRITE_ARB))) {
			/**/p_state->BindPixelPackBuffer(0);
			return false;
		}
		/**/p_state->BindPixelPackBuffer(0);
		// refresh memory mapping again
	} else {
		p_state->BindPixelUnPackBuffer(0);
		p_state->BindPixelPackBuffer(0);
		// make sure no such buffers are bound

		if(!CopyImage_To_Texture2D(p_image))
			return false;
		// copy image to texture (is bound) traditional way
	}
	// upload

	p_image_info->b_current_on_server = true;
	// image is now current on GPU

	return true;
}

/**
 *	bool Download_ImageStruct(ImageStruct *p_image)
 *		- forces download of p_image from GPU memory to system memory
 *		- returns true on success, false on failure
 */
bool Download_ImageStruct(ImageStruct *p_image) // stahne obrazek z graficke karty
{
	TImageStructInfo *p_image_info;
	if(!(p_image_info = p_FindImageInfo(p_image)))
		return false;
	// find associated texture

	if(p_image_info->b_pixel_buffer_object) {
		int n_format, n_data_type, n_internal_format;
		if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
			return false;

		_ASSERTE(p_image_info->p_pbo);
		p_image_info->p_pbo->UnMap(p_state);
		/**/p_state->BindPixelPackBuffer(0);
		// unmap buffer

		p_image_info->p_pbo->Bind(p_state);

		glPixelStorei(GL_PACK_SKIP_ROWS, 0);
		glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
		glPixelStorei(GL_PACK_ROW_LENGTH, 0);
		glPixelStorei(GL_PACK_ALIGNMENT, n_GL_PixelSize(n_internal_format));
		// set proper transfer mode

		p_image_info->p_texture->Bind_Enable(p_state);
		glGetTexImage(GL_TEXTURE_2D,
					0, // LOD pyramid level
					n_format,
					n_data_type,
					p_image_info->p_pbo->p_OffsetPointer(0));
		// fast block copy in graphics card (texture to buffer)

		//p_image_info->p_pbo->Bind(p_state); // must bind to force primary target
		if(!(p_image->Raster = (unsigned char*)p_image_info->p_pbo->p_Map(p_state,
		   GL_READ_WRITE_ARB))) {
			/**/p_state->BindPixelPackBuffer(0);
			return false;
		}
		/**/p_state->BindPixelPackBuffer(0);
		// refresh memory mapping again

		return true;
	} else {
		p_state->BindPixelUnPackBuffer(0);
		p_state->BindPixelPackBuffer(0);
		// make sure no such buffers are bound

		p_image_info->p_texture->Bind_Enable(p_state);
		return CopyTexture2D_To_Image(p_image);
		// copy data to image from currently bound texture
	}
	// download

	return true;
}

/**
 *	bool Upload_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id)
 *		- starts asynchronous upload of p_image from system memory to GPU memory
 *		  (returns almost immediately)
 *		- p_fence_id is output parameter, it will be written task identifier which can
 *		  be used in b_FenceFinished() and WaitFence() functions
 *		- in case fences aren't supported (refer to b_NVFenceSupported()), contents
 *		  of p_fence_id are undefined
 *		- returns true on success, false on failure
 */
bool Upload_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id)
{
	if(b_pixel_buffer_supported && b_FastTransfer(p_image)) {
		TImageStructInfo *p_image_info;
		if(!(p_image_info = p_FindImageInfo(p_image))) {
			if(!p_CreateImageInfo(p_image))
				return false;
			// create new associated texture, containing the image
		} else if(p_image_info->b_pixel_buffer_object) {
			if(!Upload_ImageStruct(p_image))
				return false;
			// it was mapped image. it doesn't upload this way
		} else {
			int n_format, n_data_type, n_internal_format;
			if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
				return false;
			unsigned int n_pixel_size = n_PixelSize(p_image->PixelType);
			// get pixel type and size

			if(!p_tmp_upload_buffer && !(p_tmp_upload_buffer =
			   new CGLPixelBufferObject(GL_PIXEL_UNPACK_BUFFER_ARB)))
				return false;
			// create upload buffer if necessary

			const void *p_raster = p_image->Raster;
			if(p_image->YOffset < 0)
				p_raster = (const void*)((long)p_raster + p_image->YOffset * (p_image->YSize - 1));
			if(p_image->XOffset < 0)
				p_raster = (const void*)((long)p_raster + p_image->XOffset * (p_image->XSize - 1));
			// find upper-left corner of image

#if(1)
			if(!p_tmp_upload_buffer->Alloc(p_state,
			   p_image->XSize * p_image->YSize * n_pixel_size, p_raster, GL_STREAM_DRAW_ARB)) {
				p_state->BindPixelUnPackBuffer(0);
				return false;
			}
			// start copying image data into the buffer
#else
			p_state->BindPixelUnPackBuffer(0);
			p_tmp_upload_buffer->BindAs(p_state, GL_PIXEL_PACK_BUFFER_ARB);
			if(!p_tmp_upload_buffer->Alloc(p_state,
			   p_image->XSize * p_image->YSize * n_pixel_size, 0, GL_STREAM_DRAW_ARB)) {
				return false;
			}
			// create empty buffer

			void *p_dest;
			if(!(p_dest = p_tmp_upload_buffer->p_Map(p_state, GL_WRITE_ONLY))) {
				return false;
			}
			memcpy(p_dest, p_image->Raster, p_image->XSize * p_image->YSize * n_pixel_size);
			// start copying image data into the buffer

			p_tmp_upload_buffer->UnMap(p_state);
			p_state->BindPixelPackBuffer(0);
			p_tmp_upload_buffer->Bind(p_state);
			// not-quite synchronous
#endif

			glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
			glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
			glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
			glPixelStorei(GL_UNPACK_ALIGNMENT, n_pixel_size);
			// set proper transfer mode

			p_image_info->p_texture->Bind_Enable(p_state);
			// bind destination texture

			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, p_image->XSize, p_image->YSize,
				n_format, n_data_type, p_tmp_upload_buffer->p_OffsetPointer(0));
			// tell OpenGL to source texture from PBO

			p_state->BindPixelUnPackBuffer(0);
			// unbind buffer and we're done

			p_image_info->b_current_on_server = true;
			// image is now current on GPU
		}
	} else {
		if(!Upload_ImageStruct(p_image))
			return false;
	}
	// start upload

	if(b_fence_supported) {
		glGenFencesNV(1, p_fence_id);
		glSetFenceNV(*p_fence_id, GL_ALL_COMPLETED_NV);
	}
	// create fence to determine wheter upload completed

	return true;
}

/**
 *	bool Download_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id)
 *		- starts asynchronous download of p_image from GPU memory to system memory
 *		  (returns almost immediately)
 *		- p_fence_id is output parameter, it will be written task identifier which can
 *		  be used in b_FenceFinished() and WaitFence() functions
 *		- in case fences aren't supported (refer to b_NVFenceSupported()), contents
 *		  of p_fence_id are undefined
 *		- returns true on success, false on failure
 */
bool Download_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id)
{
	bool b_result = true;
	if(b_pixel_buffer_supported) {
		TImageStructInfo *p_image_info;
		if(!(p_image_info = p_FindImageInfo(p_image)))
			return false; // image doesn't have associated texture yet
		else if(p_image_info->b_pixel_buffer_object) {
			if(!Download_ImageStruct(p_image))
				return false;
			// it was mapped image. it doesn't upload this way
		} else {
			int n_format, n_data_type, n_internal_format;
			if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
				return false;
			unsigned int n_pixel_size = n_PixelSize(p_image->PixelType);
			// get pixel type and size

			glPixelStorei(GL_PACK_SKIP_ROWS, 0);
			glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
			glPixelStorei(GL_PACK_ROW_LENGTH, 0);
			glPixelStorei(GL_PACK_ALIGNMENT, n_pixel_size);
			// set proper transfer mode

			/*if(!p_tmp_download_buffer && !(p_tmp_download_buffer =
			   new CGLPixelBufferObject(GL_PIXEL_PACK_BUFFER_ARB)))
				return false;
			// create download buffer if necessary

			p_state->BindPixelUnPackBuffer(0); // to avoid sourcing buffer data from PBO
			if(!p_tmp_download_buffer->Alloc(p_state,
			   p_image->XSize * p_image->YSize * n_pixel_size, 0, GL_STREAM_READ))
				return false;
			// create empty read-only buffer

			p_image_info->p_texture->Bind_Enable(p_state);
			glGetTexImage(GL_TEXTURE_2D, 0, n_format, n_data_type,
				p_tmp_download_buffer->p_OffsetPointer(0));
			// since buffer is read-only, this should effectively only set it's
			// data to overlap texture data and that's it

			const void *p_src_raster;
			if(!(p_src_raster = p_tmp_download_buffer->p_Map(p_state, GL_READ_ONLY_ARB)))
				return false;
			memcpy(p_image->Raster, p_src_raster, p_image->XSize * p_image->YSize * n_pixel_size);
			p_tmp_download_buffer->UnMap(p_state);
			// copy image data*/
			// older, slower, non-overlapped version

			if((!p_tmp_download_buffer && !(p_tmp_download_buffer =
			   new CGLPixelBufferObject(GL_PIXEL_PACK_BUFFER_ARB))) ||
			   (!p_tmp_download_buffer2 && !(p_tmp_download_buffer2 =
			   new CGLPixelBufferObject(GL_PIXEL_PACK_BUFFER_ARB))))
				return false;
			// create two buffers

			CGLNamedRenderBuffer_FBO *p_render_buffer;
			if(!(p_render_buffer = p_Bind_Texture_Raw(p_state, p_image_info->p_texture)))
				return false;
			// bind texture to render-buffer

			int n_half1 = p_image->YSize / 2, n_half2 = p_image->YSize - n_half1;

			if(!p_tmp_download_buffer->Bind(p_state) ||
			   !p_tmp_download_buffer->Alloc(p_state,
			   p_image->XSize * n_half1 * n_pixel_size, 0, GL_STREAM_READ)) {
				Release_Texture_Raw(p_render_buffer);
				return false;
			}
			glReadPixels(0, 0, p_image->XSize, n_half1,
				n_format, n_data_type, p_tmp_download_buffer->p_OffsetPointer(0));
			// read first half of the image to the first buffer

			if(!p_tmp_download_buffer2->Bind(p_state) ||
			   !p_tmp_download_buffer2->Alloc(p_state,
			   p_image->XSize * n_half2 * n_pixel_size, 0, GL_STREAM_READ)) {
				Release_Texture_Raw(p_render_buffer);
				return false;
			}
			glReadPixels(0, n_half1, p_image->XSize, n_half2,
				n_format, n_data_type, p_tmp_download_buffer2->p_OffsetPointer(0));
			// read second half of the image to the second buffer

			Release_Texture_Raw(p_render_buffer);
			// release render-buffer

			const unsigned char *p_src_raster;
			if(!(p_src_raster = (const unsigned char*)p_tmp_download_buffer->p_Map(p_state, GL_READ_ONLY_ARB)))
				return false;
			memcpy(p_image->Raster, p_src_raster, p_image->XSize * n_half1 * n_pixel_size);
			//
			if(!(p_src_raster = (const unsigned char*)p_tmp_download_buffer2->p_Map(p_state, GL_READ_ONLY_ARB))) {
				p_tmp_download_buffer->UnMap(p_state);
				return false;
			}
			memcpy(p_image->Raster + p_image->XSize * n_half2 * n_pixel_size, p_src_raster,
				p_image->XSize * n_half2 * n_pixel_size);
			// copy image data in two halves, before the first one is finished copying,
			// the second one is ready in memory

			p_tmp_download_buffer->UnMap(p_state);
			p_tmp_download_buffer2->UnMap(p_state);
			// unmap buffers
		}
	} else
		b_result = Download_ImageStruct(p_image);
	// start upload

	if(b_fence_supported) {
		glGenFencesNV(1, p_fence_id);
		glSetFenceNV(*p_fence_id, GL_ALL_COMPLETED_NV);
	}
	// create fence to determine wheter upload completed

	return b_result;
}

/**
 *	bool b_FenceFinished(unsigned int n_fence_id)
 *		- queries state of task with identifier n_fence_id (returns immediately)
 *		- returns true in case task has been already finished, otherwise false
 *		- always returns false in case fences aren't supported (refer to b_NVFenceSupported())
 */
bool b_FenceFinished(unsigned int n_fence_id)
{
	if(!b_fence_supported)
		return false;
	return glTestFenceNV(n_fence_id) != 0;
}

/**
 *	bool WaitFence(unsigned int n_fence_id)
 *		- waits until task with identifier n_fence_id is finished
 *		- returns false in case n_fence_id doesn't contain valid task id, otherwise true
 *		- in case fences aren't supported (refer to b_NVFenceSupported()), waits until
 *		  all pending OpenGL operations are finished
 */
bool WaitFence(unsigned int n_fence_id)
{
	if(!b_fence_supported) {
		glFlush();
		glFinish();
		return true;
	}
	if(!glIsFenceNV(n_fence_id))
		return false;
	glFinishFenceNV(n_fence_id);
	return true;
}

/**
 *	void Disable_AutoUploads()
 *		- disables automatic image uploads, meaning image is copied form system memory
 *		  to GPU memory when it's used as source parameter of some image processing function
 *		  and it's dirty flag is set to false (ie. current on client)
 *		- doesn't affect forced image transfers
 */
void Enable_AutoUploads()
{
	b_auto_uploads = true;
}

/**
 *	bool b_AutoUploads_Enabled()
 *		- returns true in case automatic uploads are enabled, otherwise false
 */
bool b_AutoUploads_Enabled()
{
	return b_auto_uploads;
}

/**
 *	void Enable_AutoDownloads()
 *		- enables automatic image downloads, meaning image is copied form GPU memory
 *		  to system memory everytime it's used as destination of some image processing function
 *		- note this is default behavior, for compatibility purposes
 */
void Enable_AutoDownloads()
{
	b_auto_downloads = true;
}

/**
 *	bool b_AutoDownloads_Enabled()
 *		- returns true in case automatic downloads are enabled, otherwise false
 */
bool b_AutoDownloads_Enabled()
{
	return b_auto_downloads;
}

/**
 *	void Disable_AutoUploads()
 *		- disables automatic image uploads, meaning image is copied form system memory
 *		  to GPU memory when it's used as source parameter of some image processing function
 *		  and it's dirty flag is set to false (ie. current on client)
 *		- doesn't affect forced image transfers
 */
void Disable_AutoUploads()
{
	b_auto_uploads = false;
}

/**
 *	void Disable_AutoDownloads()
 *		- disables automatic image downloads, meaning image is not copied form GPU memory
 *		  to system memory after used as destination of some image processing function
 *		- doesn't affect forced image transfers
 *		- note in case automatic downloads are disabled and automatic uploads enabled,
 *		  image that is rendered to and then again used in rendering will first not be
 *		  copied to it's imagestruct and then will be rewritten by old data from it's imagestruct
 */
void Disable_AutoDownloads()
{
	b_auto_downloads = false;
}

/**
 *	bool Set_DirtyFlag(const ImageStruct *p_image, bool b_current_on_server)
 *		- used to set image transfer goals in case automatic transfers are disabled
 *		- sets image dirty flag. in case b_current_on_server is true, copy of image in GPU
 *		  memory is considered valid. in case b_current_on_server is false, original image
 *		  in system memory is considered valid and shall be uploaded if needed even though
 *		  automatic uploads might be disabled
 *		- in case goal was to invalidate GPU-side copy and image has no associated texture,
 *		  returns false. otherwise returns true (in case image has no texture, it would have
 *		  to be copied anyway if needed as source to some image processing function)
 *		- doesn't affect forced image transfers
 */
bool Set_DirtyFlag(const ImageStruct *p_image, bool b_current_on_server)
{
	TImageStructInfo *p_image_info;
	if(p_image_info = p_FindImageInfo(p_image)) {
		p_image_info->b_current_on_server = b_current_on_server; // set attrib
	    return true;
	}

	return b_current_on_server == false;
	// in case we wanted to invalidate server-side copy, we haven't failed
	// (has no associated texture)
}

/**
 *	bool b_GL_ImagesSupported()
 *		- returns true in case OpenGL buffers with contents mapped to system memory are
 *		  supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension names are GL_ARB_vertex_buffer_object (promoted to core feature in
 *		  OpenGL 1.5) and GL_ARB_pixel_buffer_object (core feature in OpenGL 2.1)
 */
bool b_GL_ImagesSupported()
{
	return b_gl_images_supported;
}

/**
 *	bool b_IsGL_Image(const ImageStruct *p_image)
 *		- returns true in case p_image is image in GPU memory mapped to system memory space,
 *		  otherwise returns false (which means reading/writing from/to image raster could mean
 *		  reading/writing directly from/to GPU memory)
 *		- note address of memory mapping could be changed by operations working with this
 *		  image. refer to b_GL_Image_RelocTest().
 *		- always returns false in case p_image has no associated texture
 */
bool b_IsGL_Image(const ImageStruct *p_image)
{
	TImageStructInfo *p_image_info;
	if(p_image_info = p_FindImageInfo(p_image))
	    return p_image_info->b_pixel_buffer_object;
	return false;
}

/**
 *	bool b_GL_Image_RelocTest()
 *		- runs a simple test to find out wheter mapped buffers changes their address
 *		  in system memory while being mapped and unmapped
 *		- returns false in case it is not possible to rely on constant buffer address
 *		  (meaning constant mapped image Raster pointer), otherwise true
 *		- note this is only a test and therefore cannot be fully relied on
 */
bool b_GL_Image_RelocTest()
{
	// todo
	return true;
}

/**
 *	bool b_OpenGLError()
 *		- returns true in case OpenGL error(s) occured, otherwise false
 *		- note this clears the error flag so it wouldn't return true untill some more
 *		  OpenGL errors occured
 *		- note this calls glGetError() only
 */
bool b_OpenGLError()
{
	return glGetError() != GL_NO_ERROR;
}

/**
 *	ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format)
 *		- creates image with mapped pixel buffer. image has width n_width, height n_height
 *		  and pixel format n_format
 *		- returns pointer to a new image or NULL in case there was not enough memory or in case
 *		  mapped images aren't supported
 *		- note the Raster pointer of such an image might change accross calls to image
 *		  processing functions, using this image. refer to b_GL_Image_RelocTest()
 *		- it's Optional.Regular.Free pointer is set to function which would properly free the image
 */
ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format)
{
	int n_internal_format, n_fmt, n_data;
	if(!Get_GL_PixelType(n_format, n_fmt, n_data, n_internal_format))
		return 0;
	// get internal format

	imagestruct_list.reserve(imagestruct_list.size() + 1);
	if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
		return 0;
	// no more place in the list

	p_state->BindPixelUnPackBuffer(0);
	// explicitly bind no unpack buffer otherwise texture data would be read from there !!!

	TImageStructInfo t_temp;
	t_temp.b_current_on_server = false;
	t_temp.b_pixel_buffer_object = true;
	if(!(t_temp.p_texture = new CGLTexture_2D(p_state,
	   n_width, n_height, n_internal_format, false)))
		return 0;
	if(!t_temp.p_texture->b_Status()) {
		delete t_temp.p_texture;
		return 0;
	}
	t_temp.p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	t_temp.p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	// create texture, set clamp to edge

	//p_state->BindUnPackBuffer(0); // already bound 0 because of new texture
	// explicitly bind no unpack buffer otherwise buffer data would be read from there !!!

	if(!(t_temp.p_pbo = new CGLPixelBufferObject)) {
		delete t_temp.p_texture;
		return 0;
	}
	if(!t_temp.p_pbo->Alloc(p_state, n_PixelSize(n_format) * n_width * n_height)) {
		delete t_temp.p_texture;
		delete t_temp.p_pbo;
		return 0;
	}
	// create pixel-buffer object to be associated with texture // t_odo - don't remember how to bind

	ImageStruct *p_image_struct;
	if((p_image_struct = (ImageStruct*)malloc(sizeof(ImageStruct))) == NULL) {
		delete t_temp.p_texture;
		delete t_temp.p_pbo;
		return 0;
	}
	p_image_struct->ImageType = RegularImage;
	p_image_struct->Info = NULL;
	p_image_struct->BinInfo = NULL;
	p_image_struct->BinInfoSize = 0;
	p_image_struct->ExtraInfo = NULL;
	p_image_struct->InfoSize = 0;
	p_image_struct->References = 0;
	p_image_struct->PixelType = n_format;
	p_image_struct->XOffset = n_PixelSize(n_format);
	p_image_struct->XSize = n_width;
	p_image_struct->YOffset = n_width * n_PixelSize(n_format);
	p_image_struct->YSize = n_height;
	p_image_struct->Raster = (unsigned char*)t_temp.p_pbo->p_Map(p_state, GL_READ_WRITE_ARB);
	p_image_struct->Optional.Regular.Free = Delete_GL_Image;
	// fill imagestruct

	/**/p_state->BindPixelPackBuffer(0);

	t_temp.p_image_struct = p_image_struct;

	imagestruct_list.push_back(t_temp);
	// put it into list

	return p_image_struct;
}

static void Delete_My_Image(void *p_image)
{
	Delete_GL_Image(p_image);
	free(((ImageStruct*)p_image)->Raster);
	free(p_image);
}

/**
 *	ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format)
 *		- allocates an empty image. image has width n_width, height n_height
 *		  and pixel format n_format
 *		- returns pointer to a new image or NULL in case there was not enough memory
 *		- it's Optional.Regular.Free pointer is set to function which would properly free the image
 */
ImageStruct *p_Alloc_Image(int n_width, int n_height, short n_format)
{
	int n_internal_format, n_fmt, n_data;
	if(!Get_GL_PixelType(n_format, n_fmt, n_data, n_internal_format))
		return 0;
	// get internal format

	ImageStruct *p_image_struct;
	if((p_image_struct = (ImageStruct*)malloc(sizeof(ImageStruct))) == NULL)
		return 0;
	if((p_image_struct->Raster = (unsigned char*)malloc(n_width * n_height *
	   n_PixelSize(n_format))) == NULL) {
		free(p_image_struct);
		return 0;
	}
	p_image_struct->ImageType = RegularImage;
	p_image_struct->Info = NULL;
	p_image_struct->BinInfo = NULL;
	p_image_struct->BinInfoSize = 0;
	p_image_struct->ExtraInfo = NULL;
	p_image_struct->InfoSize = 0;
	p_image_struct->References = 0;
	p_image_struct->PixelType = n_format;
	p_image_struct->XOffset = n_PixelSize(n_format);
	p_image_struct->XSize = n_width;
	p_image_struct->YOffset = n_width * n_PixelSize(n_format);
	p_image_struct->YSize = n_height;
	p_image_struct->Optional.Regular.Free = Delete_My_Image;
	// fill imagestruct

	return p_image_struct;
}

/**
 *	void Delete_GL_Image(void *p_image)
 *		- deletes associated texture of p_image in case it has got one
 *		- in case p_image is image, created using p_Create_GL_Image, it's completely deleted,
 *		  including ImageStruct and it's Raster memory
 */
void Delete_GL_Image(void *p_gl_image)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo((ImageStruct*)p_gl_image))) != imagestruct_list.end()) {
	    (*p_image_info).Delete(p_state);
		imagestruct_list.erase(p_image_info, p_image_info + 1);
	}
	// find and destroy TImageStructInfo and OpenGL objects

	//free(p_gl_image); // (*p_image_info).Delete(); line handles it
	// free imagestruct
}

/**
 *	bool CheckBufferMapping(ImageStruct *p_image)
 *		- in case there was an error, sometimes it's necessary to check buffer mapping and
 *		  maybe renew it
 *		- returns true in case buffer data was not changed (though image data are in texture
 *		  but they might be out-of-date in case user changed image Raster contents),
 *		  returns false on failure or in case data was lost
 *		- note it's necessary to check p_image->Raster as it might be NULL if function fails
 */
bool CheckBufferMapping(ImageStruct *p_image)
{
	TImageStructInfo *p_info;
	if(!(p_info = p_FindImageInfo(p_image)) || !p_info->b_pixel_buffer_object)
		return true;
	// image isn't mapped image

	if(p_info->p_pbo->b_Mapped() && p_info->p_pbo->p_BufferPointer(p_state) == p_image->Raster) {
		/**/p_state->BindPixelPackBuffer(0);
		return true;
	}
	// buffer is mapped

	if(!p_info->p_pbo->UnMap(p_state) ||
	   !(p_image->Raster = (unsigned char*)p_info->p_pbo->p_Map(p_state, GL_READ_WRITE_ARB))) {
		delete p_info->p_pbo;
		p_state->BindPixelUnPackBuffer(0);
		// make sure no unpack buffer is bound, so Alloc doesn't source data from it!!!

		p_image->Raster = NULL;

		if(!(p_info->p_pbo = new CGLPixelBufferObject) ||
		   !p_info->p_pbo->Alloc(p_state, n_PixelSize(p_image->PixelType) *
		   p_image->XSize * p_image->YSize)) {
			/**/p_state->BindPixelPackBuffer(0);
			return false;
		}
		// create a new pixel-buffer object

		p_image->Raster = (unsigned char*)p_info->p_pbo->p_Map(p_state, GL_READ_WRITE);
		// get new raster pointer

		/**/p_state->BindPixelPackBuffer(0);

		return false;
	}
	// try to renew mapping. create a new image if fails (data is lost then)

	/**/p_state->BindPixelPackBuffer(0);

	return true;
}

/*
 *								--- ~image related functions ---
 */

/*
 *								--- image processing functions ---
 */

/**
 *	bool ImageCopy(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- copies source image to dest image. converts pixel types and
 *		  has free bilinear resampling, same as all the others
 *		- returns true on success, false on failure
 */
bool ImageCopy(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageCopy(p_dest_image, p_src_image);
		// software implementation
	}

	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image)))
		return false;
	// bind destination image as render-buffer

	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_state, p_src_image)) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		CheckBufferMapping(p_src_image);
		return false;
	}
	// bind source images as textures

	Set_NoShaderState(p_state);
	// set required GL state

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	FS_Quad_2Texcoords();
	// render fullscreen quad

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	// unbind render-buffer

	if(!Release_SrcImage(p_state, p_src_image)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	// refresh memory mapping for source images if necessary

	return true;
}

/**
 *	bool ImageImageAdd(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image addition point operator
 */
bool ImageImageAdd(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageAdd(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageFixed(p_state, p_dest_image, p_src_image0, p_src_image1, GL_ADD)) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageSub(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image subtraction point operator
 */
bool ImageImageSub(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageSub(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0,
		p_src_image1, "|subtract_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageMult(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image multiplication point operator
 */
bool ImageImageMult(ImageStruct *p_dest_image,
	ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageMult(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	bool b_cmplx0 = p_src_image0->ImageType == ImageComplex ||
		p_src_image0->ImageType == ImageComplex;
	bool b_cmplx1 = p_src_image1->ImageType == ImageComplex ||
		p_src_image1->ImageType == ImageComplex;
	// do we have complex images?

	if(!b_cmplx0 && !b_cmplx1) {
		if(!_ImageImageFixed(p_state, p_dest_image, p_src_image0, p_src_image1, GL_MODULATE)) {
			CheckBufferMapping(p_src_image0);
			CheckBufferMapping(p_src_image1);
			return false;
		}
		return true;
	}
	// used fixed-function pipeline for non-complex images

	if(!b_cmplx0 && b_cmplx1) {
		ImageStruct *p_tmp = p_src_image0;
		p_src_image0 = p_src_image1;
		p_src_image1 = p_tmp;
		b_cmplx1 = false;
		b_cmplx0 = true;
	}
	// we have shader for complex * real, not real * complex as multiplication
	// is commutative and having lots of shaders are penalized by slow switching

	_ASSERTE(b_cmplx0);
	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0,
		p_src_image1, (b_cmplx1)? "|mult_cmplx2_shader" :
		"|mult_cmplx_re_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
	// pick shader either for complex * complex or complex * real
}

/**
 *	bool ImageImageDiv(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image division point operator
 */
bool ImageImageDiv(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageDiv(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	bool b_cmplx0 = p_src_image0->ImageType == ImageComplex || p_src_image0->ImageType == ImageComplex;
	bool b_cmplx1 = p_src_image1->ImageType == ImageComplex || p_src_image1->ImageType == ImageComplex;
	// do we have complex images?

	if(!b_cmplx0 && b_cmplx1) {
		ImageStruct *p_tmp = p_src_image0;
		p_src_image0 = p_src_image1;
		p_src_image1 = p_tmp;
	}
	// we have shader for complex * real, not real * complex

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0, p_src_image1,
	   (b_cmplx0)? ((b_cmplx1)? "|divide_cmplx2_shader" : "|divide_cmplx_re_shader") :
				   ((b_cmplx1)? "|divide_re_cmplx_shader" : "|divide_shader"),
	   CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
	// pick shader either for complex / complex, complex / real, real / complex or real / real
}

class CSetBlend2ShaderParams {
protected:
	float m_f_weight0, m_f_weight1;

public:
	inline CSetBlend2ShaderParams(float f_weight0, float f_weight1)
		:m_f_weight0(f_weight0), m_f_weight1(f_weight1)
	{}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_Parameter("f_weight1").SetValue(m_f_weight0);
		r_shader.t_Find_Parameter("f_weight2").SetValue(m_f_weight1);
	}
};

/**
 *	bool ImageImageBlend(ImageStruct *p_dest_image, ImageStruct *p_src_image0,
 *		ImageStruct *p_src_image1, float f_weight0, float f_weight1)
 *		- calculates following equation: p_src_image0 * f_weight0 + p_src_image1 * f_weight1
 *		- weight magnitudes are unlimited, results are clamped based on dest image pixelformat
 */
bool ImageImageBlend(ImageStruct *p_dest_image, ImageStruct *p_src_image0,
	ImageStruct *p_src_image1, float f_weight0, float f_weight1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageBlend(p_dest_image, p_src_image0, p_src_image1, f_weight0, f_weight1);
		// software implementation
	}

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0,
	   p_src_image1, "|blend_shader2", CSetBlend2ShaderParams(f_weight0, f_weight1))) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageDecal(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- alpha-blends source image 1 onto source image 0, ie:
 *		  p_src_image0 * (1 - p_src_image1_alpha) + p_src_image1 * p_src_image1_alpha
 */
bool ImageImageDecal(ImageStruct *p_dest_image,
	ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageDecal(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageFixed(p_state, p_dest_image, p_src_image0, p_src_image1, GL_DECAL)) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageInvert(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise 1 - p_src_image
 */
bool ImageInvert(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageInvert(p_dest_image, p_src_image);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|invert_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageMin(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- pointwise minimum of p_src_image0 and p_src_image1
 */
bool ImageImageMin(ImageStruct *p_dest_image,
	ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageMin(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0,
	   p_src_image1, "|point_min_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageMin(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- pointwise maximum of p_src_image0 and p_src_image1
 */
bool ImageImageMax(ImageStruct *p_dest_image,
	ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageMax(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0,
	   p_src_image1, "|point_max_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

class CSetThresholdShaderParams {
protected:
	float m_f_thresh;

public:
	inline CSetThresholdShaderParams(float f_thresh)
		:m_f_thresh(f_thresh)
	{}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_Parameter("f_thresh").SetValue(m_f_thresh);
	}
};

/**
 *	bool ImageThresh(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_threshold)
 *		- thresholds image to f_threshold
 *		- note threshold value is relative to image pixel-type value range, for example
 *		  threshold 0.5 corresponds with both 128 for Image8 and 32767 for Image16
 *		- floating-point images are compared without any value rescaling
 */
bool ImageThresh(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_threshold)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageThresh(p_dest_image, p_src_image, f_threshold);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|threshold_shader",
	   CSetThresholdShaderParams(f_threshold))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageLog(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise log(1 + p_src_image)
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool ImageLog(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageLog(p_dest_image, p_src_image);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|log_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageExp(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise exp(p_src_image) - 1
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool ImageExp(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageExp(p_dest_image, p_src_image);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|exp_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

class CSetScaleShaderParams {
protected:
	float m_f_scale, m_f_bias;

public:
	inline CSetScaleShaderParams(float f_scale, float f_bias)
		:m_f_scale(f_scale), m_f_bias(f_bias)
	{}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_Parameter("f_scale").SetValue(m_f_scale);
		r_shader.t_Find_Parameter("f_bias").SetValue(m_f_bias);
	}
};

/**
 *	bool ImageScale(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_scale, float f_bias)
 *		- pointwise image scaling as p_src_image * f_scale + f_bias
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- scale and bias values are unlimited
 */
bool ImageScale(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_scale, float f_bias)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageScale(p_dest_image, p_src_image, f_scale, f_bias);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|scale_shader",
	   CSetScaleShaderParams(f_scale, f_bias))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

class CSetColorMatrixShaderParams {
protected:
	const float *m_p_mat4x4;
	bool m_b_transpose;

public:
	inline CSetColorMatrixShaderParams(const float *p_mat4x4, bool b_transpose)
		:m_p_mat4x4(p_mat4x4), m_b_transpose(b_transpose)
	{}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_MatrixParameter("matrix").SetValue(m_p_mat4x4, m_b_transpose);
	}
};

/**
 *	bool ImageColorMatrix(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		const float *p_mat4x4, bool b_transpose)
 *		- treats each point of p_src_image as four-dimensional vector (RGBA),
 *		  multiplies it with matrix p_mat4x4 which can be applied as is or transposed
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- matrix elements values are unlimited
 *		- for example this matrix can be used to convert RGB images to YCbCr colorspace:
 *		  (note alpha channel of such an image must contain value 1 (respectively 255))
 *
 *		float p_mat4x4[16] = { // RGB to YCbCr (ITU-R BT.601)
 *			    .299f,     .587f,     .114f,   0, // R' = .299R + .587G + .114B
 *			-.168736f, -.331264f,       .5f, .5f, // G' = .5 - .168736R - .331264G - .5B
 *			      .5f, -.418688f, -.081312f, .5f, // B' = .5 + .5R - .418688G - .081312B
 *			        0,         0,         0,   1  // A' = 1
 *		};
 */
bool ImageColorMatrix(ImageStruct *p_dest_image,
	ImageStruct *p_src_image, const float *p_mat4x4, bool b_transpose)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageColorMatrix(p_dest_image, p_src_image, p_mat4x4, b_transpose);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|color-matrix_shader",
	   CSetColorMatrixShaderParams(p_mat4x4, b_transpose))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageGreyscale(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- turns source image into greyscale, using std ".299, .587, .144" equation
 *		- in case dest is RGBA image, RGB are set to gray value, A is set to 1
 */
bool ImageGreyscale(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageGreyscale(p_dest_image, p_src_image);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image,
	   "|greyscale_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageSqrt(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise square root of p_src_image
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageLog
 */
bool ImageSqrt(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageSqrt(p_dest_image, p_src_image);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|sqrt_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

class CSetPowerShaderParams {
protected:
	float m_f_power;

public:
	inline CSetPowerShaderParams(float f_power)
		:m_f_power(f_power)
	{}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_Parameter("f_power").SetValue(m_f_power);
	}
};

/**
 *	bool ImagePow(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_power)
 *		- pointwise raises p_src_image to f_power
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageExp
 */
bool ImagePow(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_power)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImagePow(p_dest_image, p_src_image, f_power);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image,
	   "|pow_shader", CSetPowerShaderParams(f_power))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageAbs(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise absolute value of p_src_image
 */
bool ImageAbs(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageAbs(p_dest_image, p_src_image);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|abs_shader", CSetNoShaderParams())) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool _ImageAffineTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		float x0, float y0, float z0, float x1, float y1, float z1,
 *		float x2, float y2, float z2, float x3, float y3, float z3)
 *		- arbitrary image interpolation over convex quad given as (x0, y0, z0), (x1, y1, z1), ...
 *		- coordinates in OpenGL-ish fashion, orthographic projection
 *		- does not check on buffer states of images on failure
 */
bool _ImageAffineTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	float x0, float y0, float z0, float x1, float y1, float z1,
	float x2, float y2, float z2, float x3, float y3, float z3)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageAffineTransform(p_dest_image, p_src_image,
			x0, y0, z0, x1, y1, z1, x2, y2, z2, x3, y3, z3);
		// software implementation
	}

	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image)))
		return false;
	// bind destination image as render-buffer

	p_state->ActiveTextureUnit(0);
	TImageStructInfo *p_image_info;
	if(!Bind_SrcImage(p_state, p_src_image, &p_image_info)) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	// bind source images as textures

	/*p_image_info->p_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
	p_image_info->p_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST);*/

	Set_PreShaderState(p_state);
	// set required GL state

	const char *p_s_shader_name = "ps_calc_shader|bicubic_shader";

	CGLNamedShaderObject *p_shader;
	if(!(p_shader = p_GetShader(p_s_shader_name))) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	p_shader->t_Find_IntParameter("n_image_w").SetValue(p_src_image->XSize);
	p_shader->t_Find_IntParameter("n_image_h").SetValue(p_src_image->YSize);
	p_shader->Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	p_state->ClearColor4f(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT);

	glBegin(GL_QUADS);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glVertex3f(x0, y0, z0);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 0);
	glVertex3f(x1, y1, z1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 0);
	glVertex3f(x2, y2, z2);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 1);
	glVertex3f(x3, y3, z3);
	glEnd();
	// render quad with given coords

	/*p_image_info->p_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_LINEAR);
	p_image_info->p_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_LINEAR);*/

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer))
		return false;
	// unbind render-buffer

	if(!Release_SrcImage(p_state, p_src_image))
		return false;
	// refresh memory mapping for source images if necessary

	return true;
}

/**
 *	bool ImageAffineTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		float x0, float y0, float z0, float x1, float y1, float z1,
 *		float x2, float y2, float z2, float x3, float y3, float z3)
 *		- arbitrary image interpolation over convex quad given as (x0, y0, z0), (x1, y1, z1), ...
 *		- coordinates in OpenGL-ish fashion, orthographic projection
 */
bool ImageAffineTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	float x0, float y0, float z0, float x1, float y1, float z1,
	float x2, float y2, float z2, float x3, float y3, float z3)
{
	if(!_ImageAffineTransform(p_dest_image, p_src_image,
	   x0, y0, z0, x1, y1, z1, x2, y2, z2, x3, y3, z3)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3)
 *		- arbitrary image interpolation over convex quad given as (x0, y0), (x1, y1), ...
 *		- coordinates in OpenGL-ish fashion, orthographic projection
 */
bool ImageTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3)
{
	if(!_ImageAffineTransform(p_dest_image, p_src_image,
	   x0, y0, 0, x1, y1, 0, x2, y2, 0, x3, y3, 0)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageRotate(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- rotates p_src_image by f_angle in radians, clockwise
 *		- image proportions are kept intact (in contrast from ImageResize which don't mind them)
 *		- in case b_scaledown_to_fit is true, source image is scaled down so it's whole area
 *		  is visible (otherwise, rotated corners might be reaching out)
 *		- f_scale is additional scaling ratio (noscale is 1.0)
 *		- uses bicubic filter
 */
bool ImageRotate(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	float f_angle, bool b_scaledown_to_fit, float f_scale)
{
	f_angle += f_pi * .25f;
	float f_radius = (float)sqrt(2) * f_scale;
	float f_image_ratio = -f_radius * (float)p_src_image->XSize / p_dest_image->XSize;
	float f_x[4] = {(float)sin(f_angle) * f_image_ratio,
		(float)sin(f_angle + f_pi * .5f) * f_image_ratio,
		(float)sin(f_angle + f_pi) * f_image_ratio,
		(float)sin(f_angle + f_pi * 1.5f) * f_image_ratio};
	float f_y[4] = {(float)cos(f_angle) * f_radius,
		(float)cos(f_angle + f_pi * .5f) * f_radius,
		(float)cos(f_angle + f_pi) * f_radius,
		(float)cos(f_angle + f_pi * 1.5f) * f_radius};
	// calc image borders

	if(b_scaledown_to_fit) {
		float f_max = 0;
		for(int i = 0; i < 4; ++ i) {
			if(fabs(f_x[i]) > f_max)
				f_max = (float)fabs(f_x[i]);
			if(fabs(f_y[i]) > f_max)
				f_max = (float)fabs(f_y[i]);
		}
		if(f_max > 1) {
			float f_scaledown = 1 / f_max;
			for(int i = 0; i < 4; ++ i) {
				f_x[i] *= f_scaledown;
				f_y[i] *= f_scaledown;
			}
		}
	}
	// in case image corners would get clipped, scale-down image

	if(!_ImageAffineTransform(p_dest_image, p_src_image,
	   f_x[0], f_y[0], 0, f_x[1], f_y[1], 0, f_x[2], f_y[2], 0, f_x[3], f_y[3], 0)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageResize(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- resamples image p_src_image to p_dest_image
 *		- uses bicubic filter
 */
bool ImageResize(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!_ImageAffineTransform(p_dest_image, p_src_image, -1, 1, 0, -1, -1, 0, 1, -1, 0, 1, 1, 0)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

class CSet2DKernelShaderParams {
protected:
	float m_p_offset[2], m_p_origin[2];
	int m_n_kernel_w, m_n_kernel_h;

public:
	inline CSet2DKernelShaderParams(int n_width, int n_height, int n_kernel_w, int n_kernel_h)
		:m_n_kernel_w(n_kernel_w), m_n_kernel_h(n_kernel_h)
	{
		m_p_offset[0] = -1.0f / n_width;
		m_p_offset[1] = -1.0f / n_height;
		m_p_origin[0] = (n_kernel_w / -2) * m_p_offset[0];
		m_p_origin[1] = (n_kernel_h / -2) * m_p_offset[1];
	}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_Parameter("v_origin").SetValue(m_p_origin);
		r_shader.t_Find_Parameter("v_offset").SetValue(m_p_offset);
		r_shader.t_Find_IntParameter("n_kernel_w").SetValue(m_n_kernel_w);
		r_shader.t_Find_IntParameter("n_kernel_h").SetValue(m_n_kernel_h);
	}
};

/**
 *	bool ImageSquareDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageSquareDilate(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|dilate_square_shader",
	   CSet2DKernelShaderParams(p_dest_image->XSize, p_dest_image->YSize,
	   n_kernel_w, n_kernel_h))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageSquareErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageSquareErode(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|erode_square_shader",
	   CSet2DKernelShaderParams(p_dest_image->XSize, p_dest_image->YSize,
	   n_kernel_w, n_kernel_h))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageDiamondErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageDiamondErode(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|erode_diamond_shader",
	   CSet2DKernelShaderParams(p_dest_image->XSize, p_dest_image->YSize,
	   n_kernel_w, n_kernel_h))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageDiamondDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageDiamondDilate(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	if(!_ImageShader(p_state, p_dest_image, p_src_image, "|dilate_diamond_shader",
	   CSet2DKernelShaderParams(p_dest_image->XSize, p_dest_image->YSize,
	   n_kernel_w, n_kernel_h))) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageErode(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageErode(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0, p_src_image1,
	   "k2d_cs_shader|erode_image_shader", CSet2DKernelShaderParams(p_dest_image->XSize,
	   p_dest_image->YSize, p_src_image1->XSize, p_src_image1->YSize))) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageDilate(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0, p_src_image1,
	   "k2d_cs_shader|dilate_image_shader", CSet2DKernelShaderParams(p_dest_image->XSize,
	   p_dest_image->YSize, p_src_image1->XSize, p_src_image1->YSize))) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageHitMiss(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- hit-and-miss morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- structuring element is thresholded as follows: values below .25 are "zero",
 *		  values above .75 are "one", values between .25 and .75 are "don't care"
 */
bool ImageImageHitMiss(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageHitMiss(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image0, p_src_image1,
	   "k2d_cs_shader|morph_hitmiss_shader", CSet2DKernelShaderParams(p_dest_image->XSize,
	   p_dest_image->YSize, p_src_image1->XSize, p_src_image1->YSize))) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageSquareOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageSquareOpen(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	CSet2DKernelShaderParams shader_params(p_dest_image->XSize,
		p_dest_image->YSize, n_kernel_w, n_kernel_h);
	if(!_Image2PassShader(p_state, p_dest_image, p_src_image, "|erode_square_shader",
	   "|dilate_square_shader", shader_params, shader_params)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageSquareClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageSquareClose(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	CSet2DKernelShaderParams shader_params(p_dest_image->XSize,
		p_dest_image->YSize, n_kernel_w, n_kernel_h);
	if(!_Image2PassShader(p_state, p_dest_image, p_src_image, "|dilate_square_shader",
	   "|erode_square_shader", shader_params, shader_params)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageDiamondOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageDiamondOpen(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	CSet2DKernelShaderParams shader_params(p_dest_image->XSize,
		p_dest_image->YSize, n_kernel_w, n_kernel_h);
	if(!_Image2PassShader(p_state, p_dest_image, p_src_image,
	   "|erode_diamond_shader", "|dilate_diamond_shader",
	   shader_params, shader_params)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageDiamondClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageDiamondClose(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	CSet2DKernelShaderParams shader_params(p_dest_image->XSize,
		p_dest_image->YSize, n_kernel_w, n_kernel_h);
	if(!_Image2PassShader(p_state, p_dest_image, p_src_image,
	   "|dilate_diamond_shader", "|erode_diamond_shader",
	   shader_params, shader_params)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageOpen(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	CSet2DKernelShaderParams shader_params(p_dest_image->XSize,
		p_dest_image->YSize, p_src_image1->XSize, p_src_image1->YSize);
	if(!_ImageImage2PassShader(p_state, p_dest_image, p_src_image0, p_src_image1,
	   "k2d_cs_shader|erode_image_shader", "k2d_cs_shader|dilate_image_shader",
	   shader_params, shader_params)) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

/**
 *	bool ImageImageClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageClose(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image0)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image0))
				return false;
			if((p_info = p_FindImageInfo(p_src_image1)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image1))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageImageClose(p_dest_image, p_src_image0, p_src_image1);
		// software implementation
	}

	CSet2DKernelShaderParams shader_params(p_dest_image->XSize,
		p_dest_image->YSize, p_src_image1->XSize, p_src_image1->YSize);
	if(!_ImageImage2PassShader(p_state, p_dest_image, p_src_image0, p_src_image1,
	   "k2d_cs_shader|dilate_image_shader", "k2d_cs_shader|erode_image_shader",
	   shader_params, shader_params)) {
		CheckBufferMapping(p_src_image0);
		CheckBufferMapping(p_src_image1);
		return false;
	}
	return true;
}

class CSetSeparablaShaderParams {
protected:
	float m_p_offset[2], m_p_origin[2];
	int m_n_kernel_size;

public:
	inline CSetSeparablaShaderParams(int n_width, int n_height, int n_kernel_size, bool b_horizontal)
		:m_n_kernel_size(n_kernel_size)
	{
		m_p_offset[0] = (b_horizontal)? -1.0f / n_width : 0;
		m_p_offset[1] = (b_horizontal)? 0 : -1.0f / n_height;
		m_p_origin[0] = (n_kernel_size / -2) * m_p_offset[0];
		m_p_origin[1] = (n_kernel_size / -2) * m_p_offset[1];
	}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_Parameter("v_origin").SetValue(m_p_origin);
		r_shader.t_Find_Parameter("v_offset").SetValue(m_p_offset);
		r_shader.t_Find_IntParameter("n_kernel_size").SetValue(m_n_kernel_size);
	}
};

class CSetMeanShaderParams : public CSetSeparablaShaderParams {
protected:
	float m_f_scale;

public:
	inline CSetMeanShaderParams(int n_width, int n_height, int n_kernel_size, bool b_horizontal)
		:CSetSeparablaShaderParams(n_width, n_height, n_kernel_size, b_horizontal)
	{
		m_f_scale = 1.0f / n_kernel_size;
	}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		CSetSeparablaShaderParams::operator ()(r_shader);
		r_shader.t_Find_Parameter("f_scale").SetValue(m_f_scale);
	}
};

/**
 *	bool ImageMean(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- separable (two-pass) mean image filter
 *		- averaging window size is n_kernel_w  n_kernel_h pixels
 */
bool ImageMean(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageMean(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	if(!_Image2PassShader(p_state, p_dest_image, p_src_image, "|sep_mean_shader",
	   "|sep_mean_shader",
	   CSetMeanShaderParams(p_dest_image->XSize, p_dest_image->YSize, n_kernel_w, true), // horizontal pass
	   CSetMeanShaderParams(p_dest_image->XSize, p_dest_image->YSize, n_kernel_h, false))) { // vertical pass
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

class CSetMedianShaderParams {
protected:
	float m_p_offset[2];

public:
	inline CSetMedianShaderParams(float f_x_offset, float f_y_offset)
	{
		m_p_offset[0] = f_x_offset;
		m_p_offset[1] = f_y_offset;
	}

	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{
		r_shader.t_Find_Parameter("v_offset").SetValue(m_p_offset);
	}
};

/**
 *	bool ImageMedian(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_size)
 *		- separable (two-pass) median image filter
 *		- averaging window size is n_kernel_size  n_kernel_size pixels, the only supported
 *		  sizes are 3  3, 5  5, 7  7 and 15  15.
 */
bool ImageMedian(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_size)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageMedian(p_dest_image, p_src_image, n_kernel_size);
		// software implementation
	}

	const char *p_s_shader_name = 0;
	switch(n_kernel_size) {
	case 3:
		p_s_shader_name = "|sep_median3_shader";
		break;
	case 5:
		p_s_shader_name = "|sep_median5_shader";
		break;
	case 7:
		p_s_shader_name = "|sep_median7_shader";
		break;
	case 15:
		p_s_shader_name = "|sep_median15_shader";
		break;
	default:
		return false;
	}
	if(!_Image2PassShader(p_state, p_dest_image, p_src_image, p_s_shader_name, p_s_shader_name,
	   CSetMedianShaderParams(1.0f / p_dest_image->XSize, 0), // horizontal pass
	   CSetMedianShaderParams(0, 1.0f / p_dest_image->YSize))) { // vertical pass
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageLocalMin(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- finds local minimum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool ImageLocalMin(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageLocalMin(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	if(!_Image2PassShader(p_state, p_dest_image, p_src_image, "|sep_local_min_shader",
	   "|sep_local_min_shader",
	   CSetSeparablaShaderParams(p_dest_image->XSize, p_dest_image->YSize, n_kernel_w, true), // horizontal pass
	   CSetSeparablaShaderParams(p_dest_image->XSize, p_dest_image->YSize, n_kernel_h, false))) { // vertical pass
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageLocalMax(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- finds local maximum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool ImageLocalMax(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageLocalMax(p_dest_image, p_src_image, n_kernel_w, n_kernel_h);
		// software implementation
	}

	if(!_Image2PassShader(p_state, p_dest_image, p_src_image, "|sep_local_max_shader",
	   "|sep_local_max_shader",
	   CSetSeparablaShaderParams(p_dest_image->XSize, p_dest_image->YSize, n_kernel_w, true), // horizontal pass
	   CSetSeparablaShaderParams(p_dest_image->XSize, p_dest_image->YSize, n_kernel_h, false))) { // vertical pass
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool _ImageConvolveSeparable(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
 *		- separable convolution filter
 *		- note horizontal and vertical kernel doesn't have to be aligned in horizontal
 *		  or vertical direction respectively, they can both point to single image
 *		- note this handles convolution with complex kernels and / or complex image
 *		- unlike ImageConvolveSeparable doesn't check on image buffer mapping on failure
 */
bool _ImageConvolveSeparable(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
{
	int n_kernel_h_size = (p_kernel_h->XSize > p_kernel_h->YSize)?
		p_kernel_h->XSize : p_kernel_h->YSize;
	int n_kernel_v_size = (p_kernel_v->XSize > p_kernel_v->YSize)?
		p_kernel_v->XSize : p_kernel_v->YSize;
	bool b_kernel_h_complex = p_kernel_h->PixelType == ImageComplex ||
		p_kernel_h->PixelType == ImageComplexLinear;
	bool b_kernel_v_complex = p_kernel_v->PixelType == ImageComplex ||
		p_kernel_v->PixelType == ImageComplexLinear;
	bool b_src_complex = p_src_image->PixelType == ImageComplex ||
		p_src_image->PixelType == ImageComplexLinear;
	bool b_temp_tex_complex = p_dest_image->PixelType == ImageComplex ||
		p_dest_image->PixelType == ImageComplexLinear;

	// pass1 = b_src_complex * b_kernel_h_complex -> b_temp_tex_complex
	// pass2 = b_temp_tex_complex * b_kernel_v_complex -> p_dest_image

	const char *p_s_shader_names[] = {
		"k1d_cs_shader|convolution_separable_shader",				// 0 0
		"k1d_cs_shader|convolution_cmplxi_rek_separable_shader",	// 0 1
		"k1d_cs_shader|convolution_rei_cmplxk_separable_shader",	// 1 0
		"k1d_cs_shader|convolution_cmplx2_separable_shader"			// 1 1
	};
	int n_pass1_shader = ((b_src_complex)? 1 : 0) | (((b_kernel_h_complex)? 1 : 0) << 1);
	int n_pass2_shader = ((b_temp_tex_complex)? 1 : 0) | (((b_kernel_v_complex)? 1 : 0) << 1);
	// determine shader name indices for both passes

	CSetSeparablaShaderParams set_pass1_params(p_dest_image->XSize,
		p_dest_image->YSize, n_kernel_h_size, true); // horizontal pass
	CSetSeparablaShaderParams set_pass2_params(p_dest_image->XSize,
		p_dest_image->YSize, n_kernel_v_size, false); // vertical pass
	// create shader parameters for both passes

	int n_blah1, n_blah2, n_internal_format;
	if(!Get_GL_PixelType(p_dest_image->PixelType, n_blah1, n_blah2, n_internal_format))
		return false;
	// determine needed texture format

	CGLNamedTexture_2D *p_temp_texture;
	if(!(p_temp_texture = p_GetTemporaryTexture2D(TTextureFormat(p_dest_image->XSize,
	   p_dest_image->YSize, n_internal_format))))
		return false;
	// get unused / create new temporary texture

	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_Texture_Raw(p_state, p_temp_texture))) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind the texture for rendering

	p_state->ActiveTextureUnit(1);
	if(!Bind_SrcImage(p_state, p_kernel_h)) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}

	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_state, p_src_image)) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind source images as textures

	// ---- pass1 ----

	Set_PreShaderState(p_state);
	// set required GL state

	CGLNamedShaderObject *p_shader;
	if(!(p_shader = p_GetShader(p_s_shader_names[n_pass1_shader]))) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	set_pass1_params(*p_shader);
	p_shader->Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	FS_Quad_2Texcoords();
	// render fullscreen quad

	Release_Texture_Raw(p_render_buffer);
	// release texture

	p_state->ActiveTextureUnit(1);
	if(!Release_SrcImage(p_state, p_kernel_h)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// release horizontal kernel

	p_state->ActiveTextureUnit(0);
	if(!Release_SrcImage(p_state, p_src_image)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// release p_src_image0 ...

	// ---- pass2 ----

	p_temp_texture->Bind_Enable(p_state);
	// ... and use temporary texture we've just rendered into instead

	p_state->ActiveTextureUnit(1);
	if(!Bind_SrcImage(p_state, p_kernel_v)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// use vertical kernel now

	//CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image))) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind destination image as render-buffer

	if(n_pass1_shader != n_pass2_shader) {
		CGLNamedShaderObject *p_shader2;
		if(!(p_shader2 = p_GetShader(p_s_shader_names[n_pass2_shader]))) {
			Release_DestImage(p_state, p_dest_image, p_render_buffer);
			ReturnTemporaryTexture(p_temp_texture->n_Id());
			return false;
		}
		set_pass2_params(*p_shader2);
		p_shader2->Bind(p_state);
		// bind pass2 shader (lazy compile)
	} else {
		set_pass2_params(*p_shader);
		p_shader->Bind(p_state); // upload new params to GPU
	}
	// bind shader for pass2 and set it's parameters

	FS_Quad_2Texcoords();
	// render fullscreen quad

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// unbind render-buffer

	if(!Release_SrcImage(p_state, p_kernel_v)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// refresh memory mapping for source images if necessary

	p_state->ActiveTextureUnit(0);
	ReturnTemporaryTexture(p_temp_texture->n_Id());
	// release and return temporary texture back to the pool

	return true;
}

/**
 *	bool ImageConvolveSeparable(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
 *		- separable convolution filter
 *		- note horizontal and vertical kernel doesn't have to be aligned in horizontal
 *		  or vertical direction respectively, they can both point to single image
 *		- note this handles convolution with complex kernels and / or complex image
 */
bool ImageConvolveSeparable(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			if((p_info = p_FindImageInfo(p_kernel_h)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_kernel_h))
				return false;
			if((p_info = p_FindImageInfo(p_kernel_v)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_kernel_v))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageConvolveSeparable(p_dest_image, p_src_image, p_kernel_h, p_kernel_v);
		// software implementation
	}

	if(!_ImageConvolveSeparable(p_dest_image, p_src_image, p_kernel_h, p_kernel_v)) {
		CheckBufferMapping(p_src_image);
		CheckBufferMapping(p_kernel_h);
		CheckBufferMapping(p_kernel_v);
		return false;
	}
	return true;
}

/**
 *	bool ImageConvolve2D(ImageStruct *p_dest_image, ImageStruct *p_src_image, ImageStruct *p_kernel)
 *		- two-dimensional convolution filter
 *		- note this handles convolution with complex kernel and / or complex image
 */
bool ImageConvolve2D(ImageStruct *p_dest_image, ImageStruct *p_src_image, ImageStruct *p_kernel)
{
	if(!b_shaders_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			if((p_info = p_FindImageInfo(p_kernel)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_kernel))
				return false;
			// download source(s) if necessary

			if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageConvolve2D(p_dest_image, p_src_image, p_kernel);
		// software implementation
	}

	bool b_src_complex = p_src_image->PixelType == ImageComplex ||
		p_src_image->PixelType == ImageComplexLinear;
	bool b_kernel_complex = p_kernel->PixelType == ImageComplex ||
		p_kernel->PixelType == ImageComplexLinear;
	const char *p_s_shader_names[] = {
		"k2d_cs_shader|convolution_2D_shader",				// 0 0
		"k2d_cs_shader|convolution_cmplxi_rek_2D_shader",	// 0 1
		"k2d_cs_shader|convolution_rei_cmplxk_2D_shader",	// 1 0
		"k2d_cs_shader|convolution_cmplx2_2D_shader"		// 1 1
	};
	const char *p_s_shader_name = p_s_shader_names[((b_src_complex)? 1 : 0) |
		(((b_kernel_complex)? 1 : 0) << 1)];
	// determine shader name

	if(!_ImageImageShader(p_state, p_dest_image, p_src_image, p_kernel, p_s_shader_name,
	   CSet2DKernelShaderParams(p_dest_image->XSize, p_dest_image->YSize,
	   p_kernel->XSize, p_kernel->YSize))) {
		CheckBufferMapping(p_src_image);
		CheckBufferMapping(p_kernel);
		return false;
	}
	return true;
	// convolve
}

/**
 *	bool _ImageGetHistogram(ImageStruct *p_src_image, int n_histogram_width, GLenum n_color_type,
 *		unsigned int *p_hist0, unsigned int *p_hist1, unsigned int *p_hist2, unsigned int *p_hist3,
 *		GLenum n_hist0, GLenum n_hist1, GLenum n_hist2, GLenum n_hist3)
 *		- retrieves image histograms
 *		- n_histogram_width is histogram width and it must be power of two, n_color_type
 *		  specifies color components when creating histogram
 *		- p_hist0 will contain histogram of color n_hist0, the same goes for 1's, 2's and 3's
 */
bool _ImageGetHistogram(ImageStruct *p_src_image, int n_histogram_width, GLenum n_color_type,
	unsigned int *p_hist0, unsigned int *p_hist1, unsigned int *p_hist2, unsigned int *p_hist3,
	GLenum n_hist0, GLenum n_hist1, GLenum n_hist2, GLenum n_hist3)
{
	CGLPixelBufferObject imaging_pixel_buffer(GL_PIXEL_PACK_BUFFER);
	// temporary buffer object

	int n_texture_id, n_format, n_data_type;
	unsigned int n_pixel_size;
	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(n_texture_id = _n_Prepare_ImagingRendering(p_state, imaging_pixel_buffer,
	   p_render_buffer, p_src_image, n_format, n_data_type, n_pixel_size)))
		return false;
	// fill temporary imaging pixel buffer with texture and create rendering context
	// to thrash pixels into (nuisance, no pixels are written but still, we need viewport
	// large enough to accomodate our image and viewport can't be bigger than the window)

	glResetHistogram(GL_HISTOGRAM);
	glHistogram(GL_HISTOGRAM, n_histogram_width, n_color_type, GL_TRUE);
	p_state->Enable(GL_HISTOGRAM);
	// create and enable histogram, pixels sink (ie. aren't processed any further)

	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, n_pixel_size);
	// set proper transfer mode

	imaging_pixel_buffer.BindAs(p_state, GL_PIXEL_UNPACK_BUFFER);
	glDrawPixels(p_src_image->XSize, p_src_image->YSize, n_format,
		n_data_type, imaging_pixel_buffer.p_OffsetPointer(0));
	p_state->BindPixelUnPackBuffer(0);
	// draw pixels, sourcing from the buffer. pixels end up in the histogram

	glPixelStorei(GL_PACK_ALIGNMENT, 4);
	if(p_hist0)
		glGetHistogram(GL_HISTOGRAM, GL_FALSE, n_hist0, GL_UNSIGNED_INT, p_hist0);
	if(p_hist1)
		glGetHistogram(GL_HISTOGRAM, GL_FALSE, n_hist1, GL_UNSIGNED_INT, p_hist1);
	if(p_hist2)
		glGetHistogram(GL_HISTOGRAM, GL_FALSE, n_hist2, GL_UNSIGNED_INT, p_hist2);
	if(p_hist3)
		glGetHistogram(GL_HISTOGRAM, GL_FALSE, n_hist3, GL_UNSIGNED_INT, p_hist3);
	// get histogram

	p_state->Disable(GL_HISTOGRAM);
	// disable histogram

	Release_Texture_Raw(p_render_buffer);
	ReturnTemporaryTexture(n_texture_id);
	// release and return temporary texture back to the pool

	return true;
}

/**
 *	bool _ImageGetMinMax(ImageStruct *p_src_image, GLenum n_color_type,
 *		GLenum n_output_data_type, void *p_minmax0, GLenum n_minmax0)
 *		- get image minimal and maximal values of n_color_type (specifies components)
 *		- values are written to p_minmax0 in n_output_data_type format in order min, then max
 */
bool _ImageGetMinMax(ImageStruct *p_src_image, GLenum n_color_type,
	GLenum n_output_data_type, void *p_minmax0, GLenum n_minmax0)
{
	CGLPixelBufferObject imaging_pixel_buffer(GL_PIXEL_PACK_BUFFER);
	// temporary buffer object

	int n_texture_id, n_format, n_data_type;
	unsigned int n_pixel_size;
	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(n_texture_id = _n_Prepare_ImagingRendering(p_state, imaging_pixel_buffer,
	   p_render_buffer, p_src_image, n_format, n_data_type, n_pixel_size)))
		return false;
	// fill temporary imaging pixel buffer with texture and create rendering context
	// to thrash pixels into (nuisance, no pixels are written but still, we need viewport
	// large enough to accomodate our image and viewport can't be bigger than the window)

	glResetMinmax(GL_MINMAX);
	glMinmax(GL_MINMAX, n_color_type, GL_TRUE);
	p_state->Enable(GL_MINMAX);
	// create and enable minmax, pixels sink (ie. aren't processed any further)

	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, n_pixel_size);
	// set proper transfer mode

	imaging_pixel_buffer.BindAs(p_state, GL_PIXEL_UNPACK_BUFFER);
	glDrawPixels(p_src_image->XSize, p_src_image->YSize, n_format,
		n_data_type, imaging_pixel_buffer.p_OffsetPointer(0));
	p_state->BindPixelUnPackBuffer(0);
	// draw pixels, sourcing from the buffer. pixels end up in the histogram

	glPixelStorei(GL_PACK_ALIGNMENT, 4);
	glGetMinmax(GL_MINMAX, GL_TRUE, n_minmax0, n_output_data_type, p_minmax0);
	// get minmax, reset it as well

	p_state->Disable(GL_MINMAX);
	// disable minmax

	Release_Texture_Raw(p_render_buffer);
	ReturnTemporaryTexture(n_texture_id);
	// release and return temporary texture back to the pool

	return true;
}

/**
 *	bool ImageGetHistogram(ImageStruct *p_src_image, int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image histogram
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool ImageGetHistogram(ImageStruct *p_src_image, int n_histogram_width, unsigned int *p_histogram)
{
	if(!b_imaging_supported || !b_pixel_buffer_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			/*if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;*/ // no dest here
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageGetHistogram(p_src_image, n_histogram_width, p_histogram);
		// software implementation
	}

	if(!_ImageGetHistogram(p_src_image, n_histogram_width, GL_LUMINANCE,
	   p_histogram, 0, 0, 0, GL_LUMINANCE, 0, 0, 0)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
}

/**
 *	bool ImageGetLuminosityHistogram(ImageStruct *p_src_image,
 *		int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image luminosity histogram (greyscale conversion applied)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool ImageGetLuminosityHistogram(ImageStruct *p_src_image,
	int n_histogram_width, unsigned int *p_histogram)
{
	if(!b_imaging_supported || !b_pixel_buffer_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			/*if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;*/ // no dest here
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageGetLuminosityHistogram(p_src_image, n_histogram_width, p_histogram);
		// software implementation
	}

#ifdef GPU_LIB_BGR
	const GLfloat p_mat4x4[16] = {
		.114f, .114f, .114f, 0.0f, // BGRA
		.587f, .587f, .587f, 0.0f,
		.299f, .299f, .299f, 0.0f,
		0.0f,  0.0f,  0.0f,  1.0f
	};
#else //GPU_LIB_BGR
	const GLfloat p_mat4x4[16] = {
		.299f, .299f, .299f, 0.0f,
		.587f, .587f, .587f, 0.0f,
		.114f, .114f, .114f, 0.0f,
		0.0f,  0.0f,  0.0f,  1.0f
	};
#endif //GPU_LIB_BGR
    glMatrixMode(GL_COLOR);
    glLoadMatrixf(p_mat4x4);
    glMatrixMode(GL_MODELVIEW);
	// RGB to luminosity

	if(!_ImageGetHistogram(p_src_image, n_histogram_width, GL_LUMINANCE,
	   p_histogram, 0, 0, 0, GL_LUMINANCE, 0, 0, 0)) {
		glMatrixMode(GL_COLOR);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		// color matrix back

		CheckBufferMapping(p_src_image);
		return false;
	}
	glMatrixMode(GL_COLOR);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	// color matrix back

	return true;
}

/**
 *	bool ImageGetRGBHistogram(ImageStruct *p_src_image, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b)
 *		- retrieves RGB image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g and p_histo_b will contain histogram data, they doen't have
 *		  to be zeroed prior calling this
 */
bool ImageGetRGBHistogram(ImageStruct *p_src_image, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b)
{
	if(!b_imaging_supported || !b_pixel_buffer_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			/*if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;*/ // no dest here
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageGetRGBHistogram(p_src_image, n_histogram_width,
			p_histo_r, p_histo_g, p_histo_b);
		// software implementation
	}

#ifdef GPU_LIB_BGR
	if(!_ImageGetHistogram(p_src_image, n_histogram_width, GL_RGB,
	   p_histo_b, p_histo_g, p_histo_r, 0, GL_BLUE, GL_GREEN, GL_RED, 0)) { // BGRA
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
#else //GPU_LIB_BGR
	if(!_ImageGetHistogram(p_src_image, n_histogram_width, GL_RGB,
	   p_histo_b, p_histo_g, p_histo_r, 0, GL_RED, GL_GREEN, GL_BLUE, 0)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
#endif //GPU_LIB_BGR
}

/**
 *	bool ImageGetRGBAHistogram(ImageStruct *p_src_image, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g,
 *		unsigned int *p_histo_b, unsigned int *p_histo_a)
 *		- retrieves RGBA image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g, p_histo_b and p_histo_a will contain histogram data, they
 *		  doen't have to be zeroed prior calling this
 */
bool ImageGetRGBAHistogram(ImageStruct *p_src_image, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g,
	unsigned int *p_histo_b, unsigned int *p_histo_a)
{
	if(!b_imaging_supported || !b_pixel_buffer_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			/*if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;*/ // no dest here
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageGetRGBAHistogram(p_src_image, n_histogram_width,
			p_histo_r, p_histo_g, p_histo_b, p_histo_a);
		// software implementation
	}

#ifdef GPU_LIB_BGR
	if(!_ImageGetHistogram(p_src_image, n_histogram_width, GL_RGBA,
	   p_histo_b, p_histo_g, p_histo_r, p_histo_a, GL_BLUE, GL_GREEN, GL_RED, GL_ALPHA)) { // BGRA
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
#else //GPU_LIB_BGR
	if(!_ImageGetHistogram(p_src_image, n_histogram_width, GL_RGBA,
	   p_histo_b, p_histo_g, p_histo_r, p_histo_a, GL_RED, GL_GREEN, GL_BLUE, GL_ALPHA)) {
		CheckBufferMapping(p_src_image);
		return false;
	}
	return true;
#endif //GPU_LIB_BGR
}

/**
 *	bool ImageGetMinMax(ImageStruct *p_src_image, void *p_minmax)
 *		- get image minimal and maximal values
 *		- values are written to p_minmax in image pixel-type format in order min, then max
 */
bool ImageGetMinMax(ImageStruct *p_src_image, void *p_minmax)
{
	if(!b_imaging_supported || !b_pixel_buffer_supported || !b_fbo_supported) {
		if(!b_auto_uploads) {
			TImageStructInfo *p_info;
			if((p_info = p_FindImageInfo(p_src_image)) && p_info->b_current_on_server &&
			   !Download_ImageStruct(p_src_image))
				return false;
			// download source(s) if necessary

			/*if(p_info = p_FindImageInfo(p_dest_image))
				p_info->b_current_on_server = false;*/ // no dest here
			// dest is current on client
		}
		// in case auto uploads are enabled, images would upload from client anyway,
		// but otherwise versions on server would be used so we download them to acheive
		// closest possible behavior

		return rrImageGetMinMax(p_src_image, p_minmax);
		// software implementation
	}

	bool b_result;
	switch(p_src_image->PixelType) {
	case Image8:
	case Image8Linear:
		b_result = _ImageGetMinMax(p_src_image, GL_LUMINANCE,
			GL_UNSIGNED_BYTE, p_minmax, GL_LUMINANCE);
		break;
	case Image16:
	case Image16Linear:
		b_result = _ImageGetMinMax(p_src_image, GL_LUMINANCE, GL_SHORT, p_minmax, GL_LUMINANCE);
		break;
	case Image32:
	case Image32Linear:
		b_result = _ImageGetMinMax(p_src_image, GL_LUMINANCE, GL_INT, p_minmax, GL_LUMINANCE);
		break;
	case ImageFloat:
	case ImageFloatLinear:
		b_result = _ImageGetMinMax(p_src_image, GL_LUMINANCE, GL_FLOAT, p_minmax, GL_LUMINANCE);
		break;
	case ImageRGB:
	case ImageRGBLinear:
#ifdef GPU_LIB_BGR
		b_result = _ImageGetMinMax(p_src_image, GL_RGBA,
			GL_UNSIGNED_INT_8_8_8_8_REV, p_minmax, GL_BGRA);
#else //GPU_LIB_BGR
		b_result = _ImageGetMinMax(p_src_image, GL_RGBA,
			GL_UNSIGNED_INT_8_8_8_8_REV, p_minmax, GL_RGBA);
#endif //GPU_LIB_BGR
		break;
	case ImageComplex:
	case ImageComplexLinear:
		b_result = _ImageGetMinMax(p_src_image, GL_LUMINANCE_ALPHA,
			GL_FLOAT, p_minmax, GL_LUMINANCE_ALPHA);
		break;
	default:
		return false;
	}
	if(!b_result)
		CheckBufferMapping(p_src_image);
	return b_result;
}

class CFFT {
protected:
#ifdef FFT_USE_FLOAT_INDICES
	float *m_p_butterfly_index;
#else //FFT_USE_FLOAT_INDICES
	unsigned short *m_p_butterfly_index;
#endif //FFT_USE_FLOAT_INDICES
	float *m_p_butterfly_weight;
	int m_n_width;
	int m_n_log2_width;
	bool m_b_forward;

	CGLTexture_1D **m_p_index_tex_list;
	CGLTexture_1D **m_p_weight_tex_list;

public:
	/*
	 *	CFFT(int n_width)
	 *		- default constructor
	 *		- creates FFT for n_width samples for forward fft in case b_forward_fft is true,
	 *		  otherwise for reverse fft
	 *		- check b_Status() to see wheter allocation of butterfly arrays succeeded
	 */
	CFFT(int n_width, bool b_forward_fft)
		:m_n_width(n_width), m_b_forward(b_forward_fft)
	{
		m_n_log2_width = (int)(log((double)n_width) / log(2.0));
		// log2 width

#ifdef FFT_USE_FLOAT_INDICES
		m_p_butterfly_index = new float[m_n_log2_width * n_width * 2]; // two inputs
#else //FFT_USE_FLOAT_INDICES
		m_p_butterfly_index = new unsigned short[m_n_log2_width * n_width * 2]; // two inputs
#endif //FFT_USE_FLOAT_INDICES
		m_p_butterfly_weight = new float[m_n_log2_width * n_width * 2]; // im, re
		// alloc arrays

		if(b_Status()) {
			unsigned short *p_stage_i = m_p_butterfly_index;
			float *p_stage_w = m_p_butterfly_weight;
			// pointers to data of one of FFT stages (there is m_n_log2_width of them)

#ifdef FFT_USE_FLOAT_INDICES
			float f_coord_scale = 1.0f / n_width;
#endif

			for(int i = 0; i < m_n_log2_width; ++ i) {
				int n_blocks = 1 << (m_n_log2_width - 1 - i);
				int n_inputs = 1 << i;
				for(int j = 0; j < n_blocks; ++ j) {
					for(int k = 0; k < n_inputs; ++ k) {
						int n_out_1, n_out_2, n_in_1, n_in_2;
						if(!i) {
							n_out_1 = j * n_inputs * 2 + k;
							n_out_2 = j * n_inputs * 2 + n_inputs + k;
							n_in_1 = n_BitReverse(n_out_1, n_width);
							n_in_2 = n_BitReverse(n_out_2, n_width);
						} else {
							n_out_1 = j * n_inputs * 2 + k;
							n_out_2 = j * n_inputs * 2 + n_inputs + k;
							n_in_1 = n_out_1;
							n_in_2 = n_out_2;
						}
						// determine input and output indices

#ifdef FFT_USE_FLOAT_INDICES
						p_stage_i[2 * n_out_1] = n_in_1 * f_coord_scale;
						p_stage_i[2 * n_out_1 + 1] = n_in_2 * f_coord_scale;
						p_stage_i[2 * n_out_2] = n_in_1 * f_coord_scale;
						p_stage_i[2 * n_out_2 + 1] = n_in_2 * f_coord_scale;
#else //FFT_USE_FLOAT_INDICES
						p_stage_i[2 * n_out_1] = 65535 * n_in_1 / n_width;
						p_stage_i[2 * n_out_1 + 1] = 65535 * n_in_2 / n_width;
						p_stage_i[2 * n_out_2] = 65535 * n_in_1 / n_width;
						p_stage_i[2 * n_out_2 + 1] = 65535 * n_in_2 / n_width;
#endif //FFT_USE_FLOAT_INDICES
						// fill indices

						ComputeWeight(k * n_blocks, n_width, b_forward_fft,
							p_stage_w[2 * n_out_1], p_stage_w[2 * n_out_1 + 1]);
						p_stage_w[2 * n_out_2] = -p_stage_w[2 * n_out_1];
						p_stage_w[2 * n_out_2 + 1] = -p_stage_w[2 * n_out_1 + 1];
						// fill weights
					}
				}
				p_stage_i += 2 * n_width;
				p_stage_w += 2 * n_width;
			}
		}
		// fill arrays
	}

	/*
	 *	~CFFT()
	 *		- default destructor
	 */
	~CFFT()
	{
		if(m_p_butterfly_index)
			delete[] m_p_butterfly_index;
		if(m_p_butterfly_weight)
			delete[] m_p_butterfly_weight;
		if(m_p_index_tex_list) {
			for(int i = 0; i < m_n_log2_width; ++ i)
				delete m_p_index_tex_list[i];
			delete[] m_p_index_tex_list;
		}
		if(m_p_weight_tex_list) {
			for(int i = 0; i < m_n_log2_width; ++ i)
				delete m_p_weight_tex_list[i];
			delete[] m_p_weight_tex_list;
		}
	}

	/*
	 *	bool CreateTextures(CGLState *p_state)
	 *		- create textures, containing indices and weights
	 *		- returns true on success, false on failure
	 */
	bool CreateTextures(CGLState *p_state)
	{
		if(!(m_p_index_tex_list = new CGLTexture_1D*[m_n_log2_width]))
			return false;
		if(!(m_p_weight_tex_list = new CGLTexture_1D*[m_n_log2_width])) {
			delete[] m_p_index_tex_list;
			m_p_index_tex_list = 0;
			return false;
		}
		for(int i = 0; i < m_n_log2_width; ++ i) {
#ifdef FFT_USE_FLOAT_INDICES
			m_p_index_tex_list[i] = new CGLTexture_1D(p_state, m_n_width,
				GL_LUMINANCE_ALPHA16F_ARB, false, 0, GL_TEXTURE_1D, GL_LUMINANCE_ALPHA, GL_FLOAT,
				m_p_butterfly_index + (i * m_n_width * 2));
#else //FFT_USE_FLOAT_INDICES
			m_p_index_tex_list[i] = new CGLTexture_1D(p_state, m_n_width, GL_LUMINANCE16_ALPHA16,
				false, 0, GL_TEXTURE_1D, GL_LUMINANCE_ALPHA, GL_UNSIGNED_SHORT,
				m_p_butterfly_index + (i * m_n_width * 2));
#endif //FFT_USE_FLOAT_INDICES
			// create index texture (half / short)

			m_p_weight_tex_list[i] = new CGLTexture_1D(p_state, m_n_width,
				GL_LUMINANCE_ALPHA32F_ARB, false, 0, GL_TEXTURE_1D, GL_LUMINANCE_ALPHA, GL_FLOAT,
				m_p_butterfly_weight + (i * m_n_width * 2));
			// create weight texture (float)

			if(!m_p_index_tex_list[i] || !m_p_weight_tex_list[i]) {
				if(m_p_index_tex_list[i])
					delete m_p_index_tex_list[i];
				for(-- i; i >= 0; -- i) {
					delete m_p_index_tex_list[i];
					delete m_p_weight_tex_list[i];
				}
				delete[] m_p_index_tex_list;
				delete[] m_p_weight_tex_list;
				m_p_index_tex_list = 0;
				m_p_weight_tex_list = 0;

				return false;
			}
			// not enough memory?
		}

		return true;
	}

	/*
	 *	inline bool b_Status() const
	 *		- returns true in case constructor succeeded, otherwise false
	 */
	inline bool b_Status() const
	{
		return m_p_butterfly_index && m_p_butterfly_weight;
	}

	/*
	 *	inline int n_Width() const
	 *		- returns width of FFT
	 */
	inline int n_Width() const
	{
		return m_n_width;
	}

	/*
	 *	inline int n_Butterfly_Num() const
	 *		- returns number of butterflies in FFT
	 */
	inline int n_Butterfly_Num() const
	{
		return m_n_log2_width;
	}

	/*
	 *	inline bool b_Forward() const
	 *		- returns true in case FFT is forward, otherwise returns false (reverse FFT)
	 */
	inline bool b_Forward() const
	{
		return m_b_forward;
	}

	/*
	 *	inline const float *p_ButterflyIndices() const
	 *	inline const unsigned short *p_ButterflyIndices() const
	 *		- data type depends on FFT_USE_FLOAT_INDICES macro being defined
	 *		- returns array of butterfly index pairs for all butterflies in all stages
	 *		- single stage therefore has 2 * n_Width() indices
	 */
#ifdef FFT_USE_FLOAT_INDICES
	inline const float *p_ButterflyIndices() const
#else //FFT_USE_FLOAT_INDICES
	inline const unsigned short *p_ButterflyIndices() const
#endif //FFT_USE_FLOAT_INDICES
	{
		return m_p_butterfly_index;
	}

	/*
	 *	inline const float *p_ButterflyWeights() const
	 *		- returns array of complex butterfly weights for all butterflies in all stages
	 *		- single stage therefore has 2 * n_Width() elements (n_Width() complex values)
	 */
	inline const float *p_ButterflyWeights() const
	{
		return m_p_butterfly_weight;
	}

	/*
	 *	inline CGLTexture_1D *p_IndexTexture(int n_pass)
	 *		- return texture containing butterfly indices
	 */
	inline CGLTexture_1D *p_IndexTexture(int n_pass)
	{
		return m_p_index_tex_list[n_pass];
	}

	/*
	 *	inline CGLTexture_1D *p_WeightTexture(int n_pass)
	 *		- return texture containing butterfly weights
	 */
	inline CGLTexture_1D *p_WeightTexture(int n_pass)
	{
		return m_p_weight_tex_list[n_pass];
	}

protected:
	/*
	 *	static int n_BitReverse(int x, int n_bit_num)
	 *		- returns bit-reversed version of index x having n_bit_num bits
	 */
	static int n_BitReverse(int x, int n_bit_num)
	{
		int n_result = 0;
		for(int i = 0, j = n_bit_num - 1; i < n_bit_num; ++ i, -- j)
			n_result |= ((x >> i) & 1) << j;
		return n_result;
	}

	/*
	 *	static void ComputeWeight(int k, int n_width, bool b_forward_fft,
	 *		float &r_f_real, float &r_f_imag)
	 *		- compute weight for sample k, of n_width samples
	 *		- b_forward_fft affects wheter weight is for forward transform or inverse transform
	 *		- returns complex weight in r_f_real and r_f_imag
	 */
	static void ComputeWeight(int k, int n_width, bool b_forward_fft,
		float &r_f_real, float &r_f_imag)
	{
		r_f_real = (float)cos(2.0 * f_pi * k / (float)n_width);
		r_f_imag = (float)(b_forward_fft)? -sin(2.0 * f_pi * k / (float)n_width) :
			sin(2.0 * f_pi * k / (float)n_width);
	}
};

/*
 *	FFTHandle h_CreateFFT(int n_width, bool b_forward)
 *		- returns new FFT handle for n_width samples
 *		- returns InvalidFFTHandle for invalid widths (non power of two, too small)
 */
FFTHandle h_CreateFFT(int n_width, bool b_forward)
{
	CFFT *p_fft;
	if(!(p_fft = new CFFT(n_width, b_forward)))
		return InvalidFFTHandle;
	if(!p_fft->b_Status() || !p_fft->CreateTextures(p_state)) {
		delete p_fft;
		return InvalidFFTHandle;
	}
	return (FFTHandle)p_fft;
}

/*
 *	void DeleteFFT(FFTHandle h_fft)
 *		- deletes FFT handle h_fft, obtained by h_CreateFFT function
 */
void DeleteFFT(FFTHandle h_fft)
{
	CFFT *p_fft = (CFFT*)h_fft;
	delete p_fft;
}

/*
 *	bool ImageFFT(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		FFTHandle h_fft_x, FFTHandle h_fft_y)
 *		- calculates fast fourier transform of p_src_image which needs to be monochrome or complex
 *		  (color images are converted to monochrome by sampling red channel only)
 *		- h_fft_x and h_fft_y are FFT handles for horizontal and vertical pass respectively
 *		  (their size needs to match source image width and height)
 *		- source and dest image sizes must match
 */
bool ImageFFT(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	FFTHandle h_fft_x, FFTHandle h_fft_y)
{
	CFFT *p_fft_x = (CFFT*)h_fft_x;
	CFFT *p_fft_y = (CFFT*)h_fft_y;

	if(!p_fft_x || !p_fft_y || !p_fft_x->b_Status() || !p_fft_y->b_Status())
		return false;
	// invalid handles

	if(p_dest_image->XSize != p_src_image->XSize || p_src_image->XSize != p_fft_x->n_Width() ||
	   p_dest_image->YSize != p_src_image->YSize || p_src_image->YSize != p_fft_y->n_Width())
		return false;
	// sizes doesn't match

	if(p_dest_image->XSize < 2 || p_dest_image->YSize < 2)
		return false;
	// too small images

	CGLNamedTexture_2D *p_temp_texture[2];
	if(!(p_temp_texture[0] = p_GetTemporaryTexture2D(TTextureFormat(p_dest_image->XSize,
	   p_dest_image->YSize, GL_LUMINANCE_ALPHA32F_ARB))))
		return false;
	if(!(p_temp_texture[1] = p_GetTemporaryTexture2D(TTextureFormat(p_dest_image->XSize,
		p_dest_image->YSize, GL_LUMINANCE_ALPHA32F_ARB)))) {
		ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
		return false;
	}
	// get unused / create new temporary textures

	CGLNamedRenderBuffer_FBO *p_render_buffer;

	Set_PreShaderState(p_state);
	// prepare GL state

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	int n_pass = 0;
	for(int i = 0; i < p_fft_x->n_Butterfly_Num(); ++ i, ++ n_pass) {
		p_state->ActiveTextureUnit(0);
		if(!i) {
			if(!Bind_SrcImage(p_state, p_src_image)) {
				ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
				ReturnTemporaryTexture(p_temp_texture[1]->n_Id());
				return false;
			}
			// bind source image as texture
		} else {
			p_temp_texture[n_pass & 1]->Bind_Enable(p_state);
			// bind temporary texture
		}
		// bind source texture

		p_state->ActiveTextureUnit(1);
		p_state->DisableTexture2D();
		p_fft_x->p_IndexTexture(i)->Bind_Enable(p_state);
		// bind butterfly indices texture

		p_state->ActiveTextureUnit(2);
		p_state->DisableTexture2D();
		p_fft_x->p_WeightTexture(i)->Bind_Enable(p_state);
		// bind butterfly weights texture

		if(!(p_render_buffer = p_Bind_Texture_Raw(p_state, p_temp_texture[(n_pass + 1) & 1]))) {
			ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
			ReturnTemporaryTexture(p_temp_texture[1]->n_Id());
			if(!i)
				Release_SrcImage(p_state, p_src_image);
			return false;
		}
		// bind the other texture for rendering

		const char *p_s_shader_name = (i || p_src_image->PixelType == ImageComplex ||
			p_src_image->PixelType == ImageComplexLinear)? "fft_cmplx_x_shader" :
			"fft_red_x_shader";
		// use either complex shader or red shader
		
		CGLNamedShaderObject *p_shader;
		if(!(p_shader = p_GetShader(p_s_shader_name))) {
			Release_Texture_Raw(p_render_buffer);
			ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
			ReturnTemporaryTexture(p_temp_texture[1]->n_Id());
			if(!i)
				Release_SrcImage(p_state, p_src_image);
			return false;
		}
		p_shader->Bind(p_state);
		// bind shader (lazy compile)

		FS_Quad_2Texcoords();
		// render fullscreen quad

		if(!i) {
			p_state->ActiveTextureUnit(0);
			Release_SrcImage(p_state, p_src_image);
		}
		// release source image in first pass

		Release_Texture_Raw(p_render_buffer);
		// release texture it's being rendered to
	}
	// do FFT in x-direction

	for(int i = 0; i < p_fft_y->n_Butterfly_Num(); ++ i, ++ n_pass) {
		p_state->ActiveTextureUnit(0);
		p_temp_texture[n_pass & 1]->Bind_Enable(p_state);
		// bind source texture

		p_state->ActiveTextureUnit(1);
		p_state->DisableTexture2D();
		p_fft_y->p_IndexTexture(i)->Bind_Enable(p_state);
		// bind butterfly indices texture

		p_state->ActiveTextureUnit(2);
		p_state->DisableTexture2D();
		p_fft_y->p_WeightTexture(i)->Bind_Enable(p_state);
		// bind butterfly weights texture

		if(i + 1 == p_fft_y->n_Butterfly_Num())
			p_render_buffer = p_Bind_DestImage(p_state, p_dest_image);
		else
			p_render_buffer = p_Bind_Texture_Raw(p_state, p_temp_texture[(n_pass + 1) & 1]);
		if(!p_render_buffer) {
			ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
			ReturnTemporaryTexture(p_temp_texture[1]->n_Id());
			return false;
		}
		// bind the other texture (or dest image in last pass) for rendering

		CGLNamedShaderObject *p_shader;
		if(!(p_shader = p_GetShader("fft_cmplx_y_shader"))) {
			Release_DestImage(p_state, p_dest_image, p_render_buffer);
			ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
			ReturnTemporaryTexture(p_temp_texture[1]->n_Id());
			return false;
		}
		p_shader->Bind(p_state);
		// bind shader (lazy compile)

		FS_Quad_2Texcoords();
		// render fullscreen quad

		if(i + 1 == p_fft_y->n_Butterfly_Num()) {
			if(!Release_DestImage(p_state, p_dest_image, p_render_buffer)) {
				ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
				ReturnTemporaryTexture(p_temp_texture[1]->n_Id());
				return false;
			}
		} else
			Release_Texture_Raw(p_render_buffer);
		// release texture it's being rendered to
	}
	// do FFT in y-direction

	ReturnTemporaryTexture(p_temp_texture[0]->n_Id());
	ReturnTemporaryTexture(p_temp_texture[1]->n_Id());
	// return temp textures

	return true;
}

/*
 *								--- ~image processing functions ---
 */

/*
 *		-end-of-file-
 */
