/*
								+----------------------------------+
								|                                  |
								| ***  GPU image-proc library  *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|        GPU_Lib_Private.h         |
								|                                  |
								+----------------------------------+
*/

#ifndef __GPU_LIB_PRIVATE_INCLUDED
#define __GPU_LIB_PRIVATE_INCLUDED

#if defined(_MSC_VER) && !defined(__MWERKS__)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/**
 *	struct TImageStructInfo
 *		- structure, containing associated data for ImageStruct
 */
struct TImageStructInfo {
	ImageStruct *p_image_struct;
	bool b_current_on_server;
	CGLTexture_2D *p_texture;

	bool b_pixel_buffer_object;
	CGLPixelBufferObject *p_pbo; // contains image, mapped to system memory

	void Delete(CGLState *p_state)
	{
		if(p_pbo) {
			p_pbo->UnMap(p_state);
			p_state->BindPixelPackBuffer(0);
			delete p_pbo;
			p_pbo = 0;
		}
		if(p_texture) {
			delete p_texture;
			p_texture = 0;
		}
		if(p_image_struct && b_pixel_buffer_object) {
			free(p_image_struct); // allocated using malloc() in p_Create_GL_Image()
			p_image_struct = 0;
		}
	}
};

/**
 *	class CFindImageStructInfo
 *		- a simple function object for finding image info associated with imagestruct
 */
class CFindImageStructInfo {
protected:
	const ImageStruct *m_p_image_struct;

public:
	inline CFindImageStructInfo(const ImageStruct *p_image_struct)
		:m_p_image_struct(p_image_struct)
	{}

	inline bool operator ()(const TImageStructInfo &r_info)
	{
		return r_info.p_image_struct == m_p_image_struct;
	}
};

/**
 *	class CDeleteImageStructInfo
 *		- a simple function object for deleting image info's in vector
 */
class CDeleteImageStructInfo {
protected:
	CGLState *m_p_state;

public:
	inline CDeleteImageStructInfo(CGLState *p_state)
		:m_p_state(p_state)
	{}

	inline void operator ()(TImageStructInfo &r_info)
	{
		r_info.Delete(m_p_state);
	}
};

/**
 *	CGLNamedTexture_2D *p_GetTemporaryTexture2D(TTextureFormat t_format)
 *		- returns temporary 2D texture with format t_format or 0 in case error occured
 */
CGLNamedTexture_2D *p_GetTemporaryTexture2D(TTextureFormat t_format);

/**
 *	void ReturnTemporaryTexture(int n_texture_id)
 *		- returns texture n_tex_id back to circulation
 */
void ReturnTemporaryTexture(int n_texture_id);

/**
 *	CGLNamedShaderObject *p_GetShader(const char *p_s_shader_name)
 *		- gets compiled shader with name p_s_shader_name or 0 in case error occured
 */
CGLNamedShaderObject *p_GetShader(const char *p_s_shader_name);

/**
 *	CGLNamedRenderBuffer_FBO *p_GetRenderBuffer(const CGLTexture_2D *p_texture)
 *		- returns temporary render suitable buffer for rendering into 2D
 *		  texture with format of p_texture
 *		- returns 0 in case error occured
 */
CGLNamedRenderBuffer_FBO *p_GetRenderBuffer(const CGLTexture_2D *p_texture);

/**
 *	void ReturnRenderBuffer(int n_fbo_id)
 *		- returns render buffer n_fbo_id back to circulation
 */
void ReturnRenderBuffer(int n_fbo_id);

/**
 *	TImageStructInfo *p_CreateImageInfo(ImageStruct *p_image)
 *		- creates a new image info for imagestruct p_image
 *		- note it doesn't check for duplicates
 *		- returns 0 on failure
 */
TImageStructInfo *p_CreateImageInfo(ImageStruct *p_image);

/**
 *	TImageStructInfo *p_FindImageInfo(const ImageStruct *p_image)
 *		- finds existing image info for imagestruct p_image
 *		- returns 0 in case p_image doesn't have image info
 */
TImageStructInfo *p_FindImageInfo(const ImageStruct *p_image);

#endif //__GPU_LIB_PRIVATE_INCLUDED
