/*
								+----------------------------------+
								|                                  |
								|  ***  OpenGL image binding  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|          ImageBind.cpp           |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glext.h>
#include <algorithm>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLDrv.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/VertexBufferObject.h"
#include "../berLame_src/Vector.h"
#include "GPU_Lib.h"
#include "TempTex_Manager.h"
#include "Shader_Manager.h"
#include "RenderBuffer_Manager.h"
#include "GPU_Lib_Private.h"
#include "ImageTransfer.h"

/**
 *	bool Bind_SrcImage(CGLState *p_state, ImageStruct *p_image,
 *		TImageStructInfo **p_get_image_info = 0)
 *		- binds image as source to active texturing unit
 *		- p_state is OpenGL state cache, p_image is image in question
 *		- in case p_get_image_info is not 0, image info is copied there
 *		- returns true on success, false on failure
 */
bool Bind_SrcImage(CGLState *p_state, ImageStruct *p_image, TImageStructInfo **p_get_image_info)
{
	TImageStructInfo *p_image_info;
	if(!(p_image_info = p_FindImageInfo(p_image)) && !(p_image_info = p_CreateImageInfo(p_image)))
		return false;
	// make sure there is associated texture

	if(p_get_image_info)
		*p_get_image_info = p_image_info;
	// copy pointer to image info if needed

	_ASSERTE(p_image_info->p_texture);
	p_image_info->p_texture->Bind_Enable(p_state);
	// bind texture

	if(p_image_info->b_pixel_buffer_object) {
		_ASSERTE(p_image_info->p_pbo);
		p_image_info->p_pbo->UnMap(p_state);
		/**/p_state->BindPixelPackBuffer(0);
	}
	// destroy memory mapping

	if(!b_AutoUploads_Enabled() && p_image_info->b_current_on_server)
		return true;
	// in case image is current on server, do nothing

	if(p_image_info->b_pixel_buffer_object) {
		int n_format, n_data_type, n_internal_format;
		if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
			return false;

		p_image_info->p_pbo->BindAs(p_state, GL_PIXEL_UNPACK_BUFFER_ARB);

		glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
		glPixelStorei(GL_UNPACK_ALIGNMENT, n_GL_PixelSize(n_internal_format));
		// set proper transfer mode

		glTexSubImage2D(GL_TEXTURE_2D,
					0, // LOD pyramid level
					0, // x-offset
					0, // y-offset
					p_image_info->p_texture->n_Width(), // width
					p_image_info->p_texture->n_Height(), // height
					n_format,
					n_data_type,
					p_image_info->p_pbo->p_OffsetPointer(0));
		// tell OpenGL to source texture from PBO

		p_state->BindPixelUnPackBuffer(0);
	} else {
		p_state->BindPixelUnPackBuffer(0);
		p_state->BindPixelPackBuffer(0);
		// make sure no such buffers are bound

		//p_image_info->p_texture->Bind_Enable(p_state); // is bound already
		if(!CopyImage_To_Texture2D(p_image))
			return false;
		// copy image to texture (is bound) traditional way
	}

	p_image_info->b_current_on_server = true;
	// now the image is current on GPU

	return true;
}

/**
 *	bool Release_SrcImage(CGLState *p_state, ImageStruct *p_image)
 *		- releases source image
 *		- p_state is OpenGL state cache, p_image is image in question
 */
bool Release_SrcImage(CGLState *p_state, ImageStruct *p_image)
{
	TImageStructInfo *p_image_info = p_FindImageInfo(p_image);
	_ASSERTE(p_image_info); // there must be associated texture

	if(p_image_info->b_pixel_buffer_object && !p_image_info->p_pbo->b_Mapped()) {
		_ASSERTE(p_image_info->p_pbo);
		p_image_info->p_pbo->Bind(p_state); // must bind to force primary target
		if(!(p_image->Raster = (unsigned char*)p_image_info->p_pbo->p_Map(p_state,
		   GL_READ_WRITE_ARB))) {
			/**/p_state->BindPixelPackBuffer(0);
			return false;
		}
		/**/p_state->BindPixelPackBuffer(0);
	}
	// refresh memory mapping

	return true;
}

/**
 *	CGLNamedRenderBuffer_FBO *p_Bind_Texture_Raw(CGLState *p_state, CGLTexture_2D *p_texture)
 *		- binds texture p_texture for rendering
 *		- p_state is OpenGL state cache
 *		- returns render buffer which is bound to p_texture or 0 in case error occured
 */
CGLNamedRenderBuffer_FBO *p_Bind_Texture_Raw(CGLState *p_state, CGLTexture_2D *p_texture)
{
	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_GetRenderBuffer(p_texture)))
		return 0;
	// get our dreamy render buffer

	glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
	// set the right viewport

	_ASSERTE(p_texture);
	p_texture->Disable(p_state);
	// disable texture

	p_render_buffer->Bind();
	p_render_buffer->BindColorTexture_2D(*p_texture, GL_TEXTURE_2D, 0);
	// bind render-buffer and bind texture for rendering

	return p_render_buffer;
}

/**
 *	void Release_Texture_Raw(CGLNamedRenderBuffer_FBO *p_render_buffer)
 *		- ends rendering to p_render_buffer and releases bound texture
 */
void Release_Texture_Raw(CGLNamedRenderBuffer_FBO *p_render_buffer)
{
	p_render_buffer->CopyTextureData(true);
	p_render_buffer->Release();
	// release textures and render buffer (note no copying is actually made,
	// all rendering is made directly to the textures, calling it to release textures only)

	ReturnRenderBuffer(p_render_buffer->n_Id());
	// return render-buffer-object back to circulation
}

/**
 *	CGLNamedRenderBuffer_FBO *p_Bind_DestImage(CGLState *p_state, ImageStruct *p_image)
 *		- binds image as destination, returns render buffer for rendering to image's texture
 *		- p_state is OpenGL state cache, p_image is image in question
 *		- returns render buffer on success, 0 on failure
 */
CGLNamedRenderBuffer_FBO *p_Bind_DestImage(CGLState *p_state, ImageStruct *p_image)
{
	TImageStructInfo *p_image_info;
	if(!(p_image_info = p_FindImageInfo(p_image)) && !(p_image_info = p_CreateImageInfo(p_image)))
		return 0;
	// make sure there is associated texture

	if(p_image_info->b_pixel_buffer_object) {
		_ASSERTE(p_image_info->p_pbo);
		p_image_info->p_pbo->UnMap(p_state);
		/**/p_state->BindPixelPackBuffer(0);
	}
	// destroy memory mapping

	return p_Bind_Texture_Raw(p_state, p_image_info->p_texture);
}

/**
 *	bool Release_DestImage(CGLState *p_state, ImageStruct *p_image,
 *		CGLNamedRenderBuffer_FBO *p_render_buffer)
 *		- ends rendering to p_image, returns p_render_buffer
 *		- p_state is OpenGL state cache
 *		- returns true on success, false on failure
 */
bool Release_DestImage(CGLState *p_state, ImageStruct *p_image,
	CGLNamedRenderBuffer_FBO *p_render_buffer)
{
	TImageStructInfo *p_image_info = p_FindImageInfo(p_image);
	_ASSERTE(p_image_info); // there must be associated texture

	Release_Texture_Raw(p_render_buffer);

	p_image_info->b_current_on_server = true;
	// image is now current on GPU

	bool b_result;
	if(b_AutoDownloads_Enabled()) {
		if(p_image_info->b_pixel_buffer_object) {
			p_image_info->p_pbo->Bind(p_state);

			int n_format, n_data_type, n_internal_format;
			b_result = Get_GL_PixelType(p_image->PixelType,
				n_format, n_data_type, n_internal_format);

			glPixelStorei(GL_PACK_SKIP_ROWS, 0);
			glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
			glPixelStorei(GL_PACK_ROW_LENGTH, 0);
			glPixelStorei(GL_PACK_ALIGNMENT, n_GL_PixelSize(n_internal_format));
			// set proper transfer mode

			p_image_info->p_texture->Bind_Enable(p_state);
			glGetTexImage(GL_TEXTURE_2D,
						0, // LOD pyramid level
						n_format,
						n_data_type,
						p_image_info->p_pbo->p_OffsetPointer(0));
			// fast block copy in graphics card (texture to buffer)
		} else {
			p_state->BindPixelUnPackBuffer(0);
			p_state->BindPixelPackBuffer(0);
			// make sure no such buffers are bound

			p_image_info->p_texture->Bind_Enable(p_state);
			b_result = CopyTexture2D_To_Image(p_image);
			// copy data to image from currently bound texture
		}
	} else
		b_result = true;

	if(p_image_info->b_pixel_buffer_object) {
		_ASSERTE(p_image_info->p_pbo);
		p_image_info->p_pbo->Bind(p_state); // must bind to force primary target
		if(!(p_image->Raster = (unsigned char*)p_image_info->p_pbo->p_Map(p_state,
		   GL_READ_WRITE_ARB))) {
			/**/p_state->BindPixelPackBuffer(0);
			return false;
		}
		/**/p_state->BindPixelPackBuffer(0);
	}
	// refresh memory mapping

	return b_result;
}

/*
 *		-end-of-file-
 */
