/*
								+----------------------------------+
								|                                  |
								|    ***  Image processing  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|           ImageProc.h            |
								|                                  |
								+----------------------------------+
*/

#ifndef __IMAGE_PROCESSING_INCLUDED
#define __IMAGE_PROCESSING_INCLUDED

/**
 *	inline void FS_Quad_1Texcoord()
 *		- draws fullscreen quad (or triangle, depending on GPU_LIB_USE_FULLSCREEN_TRIANGLE
 *		  directive) with texture coordinates for texture unit 0
 */
inline void FS_Quad_1Texcoord()
{
#ifndef GPU_LIB_USE_FULLSCREEN_TRIANGLE
	glBegin(GL_QUADS);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glVertex2f(-1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 0);
	glVertex2f(-1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 0);
	glVertex2f( 1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 1);
	glVertex2f( 1,  1);
	glEnd();
#else //GPU_LIB_USE_FULLSCREEN_TRIANGLE
	glBegin(GL_TRIANGLES);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glVertex2f(-1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, -1);
	glVertex2f(-1, -3);
	glMultiTexCoord2f(GL_TEXTURE0, 2, 1);
	glVertex2f( 3,  1);
	glEnd();
#endif //GPU_LIB_USE_FULLSCREEN_TRIANGLE
}

/**
 *	inline void FS_Quad_2Texcoords()
 *		- draws fullscreen quad (or triangle, depending on GPU_LIB_USE_FULLSCREEN_TRIANGLE
 *		  directive) with texture coordinates for texture units 0 and 1
 */
inline void FS_Quad_2Texcoords()
{
#ifndef GPU_LIB_USE_FULLSCREEN_TRIANGLE
	glBegin(GL_QUADS);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 1);
	glVertex2f(-1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 0);
	glVertex2f(-1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 0);
	glVertex2f( 1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 1);
	glVertex2f( 1,  1);
	glEnd();
#else //GPU_LIB_USE_FULLSCREEN_TRIANGLE
	glBegin(GL_TRIANGLES);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 1);
	glVertex2f(-1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, -1);
	glMultiTexCoord2f(GL_TEXTURE1, 0, -1);
	glVertex2f(-1, -3);
	glMultiTexCoord2f(GL_TEXTURE0, 2, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 2, 1);
	glVertex2f( 3,  1);
	glEnd();
#endif //GPU_LIB_USE_FULLSCREEN_TRIANGLE
}

/**
 *	inline void Set_NoShaderState(CGLState *p_state)
 *		- sets OpenGL state for rendering using fixed pipeline
 *		- p_state is OpenGL state cache
 */
inline void Set_NoShaderState(CGLState *p_state)
{
	p_state->DisableDepthTest();
	p_state->DisableFog();
	p_state->DisableLighting();
	p_state->DisableAllClipPlanes();
	p_state->DisableAlphaTest();
	p_state->DisableFragmentProgram();
	p_state->DisableStencilTest();
	p_state->DisableVertexProgram();
	if(p_state->p_State()->b_shader_support)
		p_state->BindProgramObject(0);
	if(p_state->p_State()->b_shader_gl2_support)
		p_state->BindProgramObject_GL2(0);
	p_state->Disable_STRQ_Texgen();
	p_state->DisableBlend();
	p_state->Color3f(1, 1, 1);
}

/**
 *	inline void Set_PreShaderState(CGLState *p_state)
 *		- sets OpenGL state for rendering using shader
 *		- p_state is OpenGL state cache
 */
inline void Set_PreShaderState(CGLState *p_state)
{
	p_state->DisableDepthTest();
	p_state->DisableAllClipPlanes();
	p_state->DisableAlphaTest();
	p_state->DisableFragmentProgram();
	p_state->DisableStencilTest();
	p_state->DisableVertexProgram();
	p_state->Disable_STRQ_Texgen();
	p_state->DisableBlend();
}

/**
 *	bool _ImageImageFixed(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
 *		ImageStruct *p_src_image1, GLenum n_image1_env_mode)
 *		- processes two images p_src_image0, p_src_image1 using fixed pipeline
 *		- result is rendered to p_dest_image
 *		- p_state is OpenGL state cache
 *		- second texture unit is set to n_image1_env_mode
 *		- returns true on success, false on failure
 */
bool _ImageImageFixed(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
	ImageStruct *p_src_image1, GLenum n_image1_env_mode);

/**
 *	int _n_Prepare_ImagingRendering(CGLState *p_state, CGLPixelBufferObject &r_tmp_imaging_buffer,
 *		CGLNamedRenderBuffer_FBO *&r_p_render_buffer, ImageStruct *p_src_image, int &r_n_format,
 *		int &r_n_data_type, unsigned int &r_n_pixel_size)
 *		- prepares imaging rendering which needs to have data in memory area, accessible for
 *		  glDrawPixels(). copies data from p_src_image to r_tmp_imaging_buffer,
 *		  creates r_p_render_buffer for rendering p_src_image (needs it even if pixels are sinked)
 *		  and fills texture format informations
 *		- returns texture id (texture name for manager, not OpenGL id) or -1 on failure
 */
int _n_Prepare_ImagingRendering(CGLState *p_state, CGLPixelBufferObject &r_tmp_imaging_buffer,
	CGLNamedRenderBuffer_FBO *&r_p_render_buffer, ImageStruct *p_src_image, int &r_n_format,
	int &r_n_data_type, unsigned int &r_n_pixel_size);

/**
 *	class CSetNoShaderParams
 *		- empty function object for setting shader parameters
 */
class CSetNoShaderParams {
public:
	inline void operator ()(CGLNamedShaderObject &r_shader) const
	{}
};

/**
 *	template <class CSetShaderParams>
 *	bool _ImageShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		const char *p_s_shader_name, CSetShaderParams set_params)
 *		- processes single image p_src_image using shader p_s_shader_name
 *		- result is rendered to p_dest_image
 *		- p_state is OpenGL state cache
 *		- set_params is function object for setting up miscelaneous shader params
 *		  (can be CSetNoShaderParams instance)
 *		- returns true on success, false on failure
 */
template <class CSetShaderParams>
bool _ImageShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image,
	const char *p_s_shader_name, CSetShaderParams set_params)
{
	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image)))
		return false;
	// bind destination image as render-buffer

	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_state, p_src_image)) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	// bind source images as textures

	Set_PreShaderState(p_state);
	// set required GL state

	CGLNamedShaderObject *p_shader;
	if(!(p_shader = p_GetShader(p_s_shader_name))) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	set_params(*p_shader);
	p_shader->Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	FS_Quad_1Texcoord();
	// render fullscreen quad

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer))
		return false;
	// unbind render-buffer

	if(!Release_SrcImage(p_state, p_src_image))
		return false;
	// refresh memory mapping for source images if necessary

	return true;
}

/**
 *	template <class CSetShaderParams>
 *	bool _ImageImageShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
 *		ImageStruct *p_src_image1, const char *p_s_shader_name, CSetShaderParams set_params)
 *		- processes two images p_src_image0 and p_src_image1 using shader p_s_shader_name
 *		- result is rendered to p_dest_image
 *		- p_state is OpenGL state cache
 *		- set_params is function object for setting up miscelaneous shader params
 *		  (can be CSetNoShaderParams instance)
 *		- returns true on success, false on failure
 */
template <class CSetShaderParams>
bool _ImageImageShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
	ImageStruct *p_src_image1, const char *p_s_shader_name, CSetShaderParams set_params)
{
	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image)))
		return false;
	// bind destination image as render-buffer

	p_state->ActiveTextureUnit(1);
	if(!Bind_SrcImage(p_state, p_src_image1)) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_state, p_src_image0)) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	// bind source images as textures

	Set_PreShaderState(p_state);
	// set required GL state

	CGLNamedShaderObject *p_shader;
	if(!(p_shader = p_GetShader(p_s_shader_name))) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	set_params(*p_shader);
	p_shader->Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	FS_Quad_2Texcoords();
	// render fullscreen quad

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer))
		return false;
	// unbind render-buffer

	p_state->ActiveTextureUnit(1);
	if(!Release_SrcImage(p_state, p_src_image1))
		return false;
	p_state->ActiveTextureUnit(0);
	if(!Release_SrcImage(p_state, p_src_image0))
		return false;
	// refresh memory mapping for source images if necessary

	return true;
}

/**
 *	template <class CSetShaderParams>
 *	bool _ImageImage2PassShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
 *		ImageStruct *p_src_image1, const char *p_s_pass1_shader_name, const char *p_s_pass2_shader_name,
 *		CSetShaderParams set_pass1_params, CSetShaderParams set_pass2_params)
 *		- processes images p_src_image0 and p_src_image1 using shader p_s_shader_name in first
 *		  pass and shader p_s_pass2_shader_name in the second pass (can be equal); in second pass
 *		  results of first pass are used instead of p_src_image1
 *		- result is rendered to p_dest_image, uses temporary texture which it gets from manager
 *		- p_state is OpenGL state cache
 *		- set_pass1_params and set_pass2_params are function objects for setting up miscelaneous
 *		  shader params (can be CSetNoShaderParams instance)
 *		- returns true on success, false on failure
 */
template <class CSetShaderParams>
bool _ImageImage2PassShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
	ImageStruct *p_src_image1, const char *p_s_pass1_shader_name, const char *p_s_pass2_shader_name,
	CSetShaderParams set_pass1_params, CSetShaderParams set_pass2_params)
{
	int n_blah1, n_blah2, n_internal_format;
	if(!Get_GL_PixelType(p_dest_image->PixelType, n_blah1, n_blah2, n_internal_format))
		return false;
	// determine needed texture format

	CGLNamedTexture_2D *p_temp_texture;
	if(!(p_temp_texture = p_GetTemporaryTexture2D(TTextureFormat(p_dest_image->XSize,
	   p_dest_image->YSize, n_internal_format))))
		return false;
	// get unused / create new temporary texture

	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_Texture_Raw(p_state, p_temp_texture))) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind the texture for rendering

	p_state->ActiveTextureUnit(1);
	if(!Bind_SrcImage(p_state, p_src_image1)) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_state, p_src_image0)) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind source images as textures

	// ---- pass1 ----

	Set_PreShaderState(p_state);
	// set required GL state

	CGLNamedShaderObject *p_shader;
	if(!(p_shader = p_GetShader(p_s_pass1_shader_name))) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	set_pass1_params(*p_shader);
	p_shader->Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	FS_Quad_2Texcoords(); // upside down
	// render fullscreen quad

	Release_Texture_Raw(p_render_buffer);
	// release texture

	p_state->ActiveTextureUnit(0);
	if(!Release_SrcImage(p_state, p_src_image0)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// release p_src_image0 ...

	// ---- pass2 ----

	p_temp_texture->Bind_Enable(p_state);
	// ... and use temporary texture we've just rendered into instead

	//CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image))) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind destination image as render-buffer

	if(p_s_pass1_shader_name != p_s_pass2_shader_name && // mostly static strings used
	   strcmp(p_s_pass1_shader_name, p_s_pass2_shader_name)) {
		CGLNamedShaderObject *p_shader2;
		if(!(p_shader2 = p_GetShader(p_s_pass2_shader_name))) {
			Release_DestImage(p_state, p_dest_image, p_render_buffer);
			ReturnTemporaryTexture(p_temp_texture->n_Id());
			return false;
		}
		set_pass2_params(*p_shader2);
		p_shader2->Bind(p_state);
		// bind pass2 shader (lazy compile)
	} else {
		set_pass2_params(*p_shader);
		p_shader->Bind(p_state); // upload new params to GPU
	}
	// bind shader for pass2 and set it's parameters

	FS_Quad_2Texcoords();
	// render fullscreen quad

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// unbind render-buffer

	ReturnTemporaryTexture(p_temp_texture->n_Id());
	// release and return temporary texture back to the pool

	p_state->ActiveTextureUnit(1);
	if(!Release_SrcImage(p_state, p_src_image1))
		return false;
	// refresh memory mapping for source images if necessary

	return true;
}

/**
 *	template <class CSetShaderParams>
 *	bool _Image2PassShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		const char *p_s_pass1_shader_name, const char *p_s_pass2_shader_name,
 *		CSetShaderParams set_pass1_params, CSetShaderParams set_pass2_params)
 *		- processes single image p_src_image using shader p_s_shader_name in first pass and
 *		  shader p_s_pass2_shader_name in the second pass (can be equal)
 *		- result is rendered to p_dest_image, uses temporary texture which it gets from manager
 *		- p_state is OpenGL state cache
 *		- set_pass1_params and set_pass2_params are function objects for setting up miscelaneous
 *		  shader params (can be CSetNoShaderParams instance)
 *		- returns true on success, false on failure
 */
template <class CSetShaderParams>
bool _Image2PassShader(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image,
	const char *p_s_pass1_shader_name, const char *p_s_pass2_shader_name,
	CSetShaderParams set_pass1_params, CSetShaderParams set_pass2_params)
{
	int n_blah1, n_blah2, n_internal_format;
	if(!Get_GL_PixelType(p_dest_image->PixelType, n_blah1, n_blah2, n_internal_format))
		return false;
	// determine needed texture format

	CGLNamedTexture_2D *p_temp_texture;
	if(!(p_temp_texture = p_GetTemporaryTexture2D(TTextureFormat(p_dest_image->XSize,
	   p_dest_image->YSize, n_internal_format))))
		return false;
	// get unused / create new temporary texture

	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_Texture_Raw(p_state, p_temp_texture))) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind the texture for rendering

	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_state, p_src_image)) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind source image as texture

	// ---- pass1 ----

	Set_PreShaderState(p_state);
	// set required GL state

	CGLNamedShaderObject *p_shader;
	if(!(p_shader = p_GetShader(p_s_pass1_shader_name))) {
		Release_Texture_Raw(p_render_buffer);
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	set_pass1_params(*p_shader);
	p_shader->Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	FS_Quad_1Texcoord(); // upside down
	// render fullscreen quad

	Release_Texture_Raw(p_render_buffer);
	// release texture

	if(!Release_SrcImage(p_state, p_src_image)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// release p_src_image ...

	// ---- pass2 ----

	p_temp_texture->Bind_Enable(p_state);
	// ... and use temporary texture we've just rendered into instead

	//CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image))) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// bind destination image as render-buffer

	if(p_s_pass1_shader_name != p_s_pass2_shader_name && // mostly static strings used
	   strcmp(p_s_pass1_shader_name, p_s_pass2_shader_name)) {
		CGLNamedShaderObject *p_shader;
		if(!(p_shader = p_GetShader(p_s_pass2_shader_name))) {
			Release_DestImage(p_state, p_dest_image, p_render_buffer);
			ReturnTemporaryTexture(p_temp_texture->n_Id());
			return false;
		}
		set_pass2_params(*p_shader);
		p_shader->Bind(p_state);
		// bind pass2 shader (lazy compile)
	} else {
		set_pass2_params(*p_shader);
		p_shader->Bind(p_state); // upload new params to GPU
	}
	// bind shader for pass2 and set it's parameters

	FS_Quad_2Texcoords();
	// render fullscreen quad

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer)) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return false;
	}
	// unbind render-buffer

	ReturnTemporaryTexture(p_temp_texture->n_Id());
	// release and return temporary texture back to the pool

	return true;
}

#endif //__IMAGE_PROCESSING_INCLUDED
