/*
								+----------------------------------+
								|                                  |
								|   *** Renderbuffer manager ***   |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|            Manager.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __GL_RENDERBUFFER_MANAGER_INCLUDED
#define __GL_RENDERBUFFER_MANAGER_INCLUDED

#include "Manager.h"

#include "TempTex_Manager.h" // TTextureFormat

/**
 *	class CGLNamedRenderBuffer_FBO
 *		- extension of render-buffer with automatically assigned id
 */
class CGLNamedRenderBuffer_FBO : public CGLRenderBuffer_FBO {
protected:
	static int m_n_id_space;
	int m_n_id;

public:
	inline int n_Id() const { return m_n_id; }
	inline GLenum n_Pixel_Format() const { return m_n_color_internal_format; }

	inline CGLNamedRenderBuffer_FBO(int n_width, int n_height,
		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
		:CGLRenderBuffer_FBO(n_width, n_height,
		 n_color_buffer_num, b_color_texture_target, n_color_internal_format,
		 b_depth_buffer, b_depth_texture_target, n_depth_internal_format,
		 b_stencil_buffer, b_stencil_texture_target, n_stencil_internal_format)
	{
		m_n_id = m_n_id_space ++;
	}
};

//int CGLNamedRenderBuffer_FBO::m_n_id_space = 0;
// needs to be declared once

/**
 *	class CGetFBOCost
 *		- cost evaluator for render-buffer manager purging purposes
 *		- currently returns size of buffer surface in pixels
 */
class CGetFBOCost {
public:
	inline int n_Cost(const CGLNamedRenderBuffer_FBO &r_fbo) const
	{
		return r_fbo.n_Width() * r_fbo.n_Height(); // todo - calculate with pixel size
	}
};

/**
 *	class CFindRB_Format
 *		- finds render-buffer by 2d-texture it's necessary to be rendered to
 */
class CFindRB_Format {
protected:
	TTextureFormat m_t_format;

public:
	inline CFindRB_Format(const TTextureFormat &r_format)
		:m_t_format(r_format)
	{}

	inline bool operator ()(const TTextureFormat &r_fbo) const
	{
		return r_fbo.n_width == m_t_format.n_width &&
			r_fbo.n_height == m_t_format.n_height &&
			r_fbo.n_internal_format == m_t_format.n_internal_format;
	}
};

/**
 *	class CRenderBufferFactory
 *		- render-buffer factory
 */
class CRenderBufferFactory {
public:
	inline CGLNamedRenderBuffer_FBO *Load(const TTextureFormat &r_format)
	{
		return new CGLNamedRenderBuffer_FBO(r_format.n_width, r_format.n_height,
			1, true, r_format.n_internal_format, false, false, 0, false, false, 0);
	}

	static inline bool b_Valid_Object(const CGLNamedRenderBuffer_FBO *p_object)
	{
		return p_object != 0;
	}
};

/**
 *	class CGLRenderBuffer_Manager
 *		- OpenGL render buffer manager class
 */
typedef CResourceManager<CGLNamedRenderBuffer_FBO, CGetFBOCost, int, CFindIntId<CGLNamedRenderBuffer_FBO>,
	TTextureFormat, CFindRB_Format, CRenderBufferFactory> CGLRenderBuffer_Manager;

#endif //__GL_RENDERBUFFER_MANAGER_INCLUDED
