/*
								+---------------------------------+
								|                                 |
								|      ***   OpenGL2.0   ***      |
								|                                 |
								|  Copyright   -tHE SWINe- 2005  |
								|                                 |
								|         RenderBuffer.cpp        |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 10-05-2006
 *
 *	integers, containing just true and false were replaced with bool
 *	fixed some design issues to enable binding depth and stencil output texture along with
 *	color texture
 *	WGL_ARB_pixel_buffer support was deprecated
 *	files were renamed to RenderBuffer2.cpp and RenderBuffer2.h respectively
 *
 *	03-06-2007
 *
 *	removed __declspec(dllexport)
 *
 */

#include <vector>
#include "../berLame_src/CallStack.h"
#include <windows.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/wglext.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "Texture.h"
#include "RenderBuffer2.h"

/*
 *								=== CGLRenderBuffer ===
 */

CGLRenderBuffer::CGLRenderBuffer(int n_width, int n_height,
	int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
	bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
	bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
	:m_p_state(0), m_b_active(false), m_b_status(false), m_n_bound_color_texture_num(0),
	m_n_bound_depth_texture(0), m_n_bound_stencil_texture(0),
	m_n_width(n_width), m_n_height(n_height),
	m_n_draw_buffer_num(n_color_buffer_num), m_n_color_internal_format(n_color_internal_format),
	m_b_depth_buffer(b_depth_buffer), m_n_depth_internal_format(n_depth_internal_format),
	m_b_stencil_buffer(b_stencil_buffer), m_n_stencil_internal_format(n_stencil_internal_format),
	m_b_color_texture_target(b_color_texture_target),
	m_b_depth_texture_target(b_depth_texture_target),
	m_b_stencil_texture_target(b_stencil_texture_target)
{
	__FuncGuard("CGLRenderBuffer::CGLRenderBuffer");

	memset(m_p_bound_color_texture, 0, sizeof(m_p_bound_color_texture));
}

CGLRenderBuffer::CGLRenderBuffer(int n_width, int n_height,
	int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_format, int n_color_bit_num, bool b_float_color,
	bool b_depth_buffer, bool b_depth_texture_target, int n_depth_bit_num,
	bool b_stencil_buffer, bool b_stencil_texture_target, int n_stencil_bit_num)
	:m_p_state(0), m_b_active(false), m_b_status(false), m_n_bound_color_texture_num(0),
	m_n_bound_depth_texture(0), m_n_bound_stencil_texture(0),
	m_n_width(n_width), m_n_height(n_height),
	m_n_draw_buffer_num(n_color_buffer_num),
	m_b_depth_buffer(b_depth_buffer),
	m_b_stencil_buffer(b_stencil_buffer),
	m_b_color_texture_target(b_color_texture_target),
	m_b_depth_texture_target(b_depth_texture_target),
	m_b_stencil_texture_target(b_stencil_texture_target)
{
	__FuncGuard("CGLRenderBuffer::CGLRenderBuffer");

	if(b_depth_buffer) {
		switch(n_depth_bit_num) {
		case 16:
			m_n_depth_internal_format = GL_DEPTH_COMPONENT16;
			break;
		case 24:
			m_n_depth_internal_format = GL_DEPTH_COMPONENT24;
			break;
		case 32:
			m_n_depth_internal_format = GL_DEPTH_COMPONENT32;
			break;
		default:
			m_n_depth_internal_format = 0;
			m_b_status = false;
			return;
		}
	}
	// find depth component format

	if(b_stencil_buffer) {
		switch(n_stencil_bit_num) {
		case 1:
			m_n_depth_internal_format = GL_STENCIL_INDEX1_EXT;
			break;
		case 4:
			m_n_depth_internal_format = GL_STENCIL_INDEX4_EXT;
			break;
		case 8:
			m_n_depth_internal_format = GL_STENCIL_INDEX8_EXT;
			break;
		case 16:
			m_n_depth_internal_format = GL_STENCIL_INDEX16_EXT;
			break;
		default:
			m_n_depth_internal_format = 0;
			m_b_status = false;
			return;
		}
	}
	// find stencil component format

	if(n_color_buffer_num) {
		if(!(m_n_color_internal_format = n_InternalColorFormat(n_color_format,
		   n_color_bit_num, b_float_color))) {
			m_b_status = false;
			return;
		}
	}
	// find color component format

	memset(m_p_bound_color_texture, 0, sizeof(m_p_bound_color_texture));
}

/*
 *	static bool CGLRenderBuffer::b_Supported()
 *		- returns true in case the render buffer class have necessary hardware support
 */
bool CGLRenderBuffer::b_Supported()
{
	__FuncGuard("CGLRenderBuffer::b_Supported");

	return false; // this is virtual class only
}

/*
 *	static int CGLRenderBuffer::n_Max_DrawBuffer_Num()
 *		- returns maximal number of OpenGL draw buffers (1 where not supported, otherwise
 *		  1, 2, 4, up to 16 in future)
 */
int CGLRenderBuffer::n_Max_DrawBuffer_Num()
{
	__FuncGuard("CGLRenderBuffer::n_Max_DrawBuffer_Num");

	return 0; // this is virtual class only
}

/*
 *	virtual bool CGLRenderBuffer::b_Status() const
 *		- returns true in case render buffer, specified by constructor parameters
 *		  was successfuly created and is ready to use
 */
bool CGLRenderBuffer::b_Status() const
{
	__FuncGuard("CGLRenderBuffer::b_Status");

	return m_b_status;
}

/*
 *	int CGLRenderBuffer::n_Width() const
 *		- returns width of the render buffer
 */
int CGLRenderBuffer::n_Width() const
{
	__FuncGuard("CGLRenderBuffer::n_Width");

	return m_n_width;
}

/*
 *	int CGLRenderBuffer::n_Height() const
 *		- returns height of the render buffer
 */
int CGLRenderBuffer::n_Height() const
{
	__FuncGuard("CGLRenderBuffer::n_Height");

	return m_n_height;
}

/*
 *	static GLenum CGLRenderBuffer::n_InternalColorFormat(GLenum n_color_format,
 *		int n_bit_num, bool b_float)
 *		- n_color_format is one of GL_ALPHA, GL_LUMINANCE, GL_LUMINANCE_ALPHA, GL_INTENSITY,
 *		  GL_RGB or GL_RGBA
 *		- n_bit_num can be 4, 8 or 16 for integer and 16 or 32 for float representation
 *		- in case b_float is true, float representation will be used. otherwise integer
 *		  representation will be used
 *		- returns OpenGL internal color format (or 0 in case no specified format exists)
 */
GLenum CGLRenderBuffer::n_InternalColorFormat(GLenum n_color_format,
	int n_bit_num, bool b_float)
{
	__FuncGuard("CGLRenderBuffer::n_InternalColorFormat");

	const GLenum p_format_array[][4] = {
		{GL_ALPHA4, GL_ALPHA8, GL_ALPHA16, 0},
		{GL_LUMINANCE4, GL_LUMINANCE8, GL_LUMINANCE16, 0},
		{GL_LUMINANCE4_ALPHA4, GL_LUMINANCE8_ALPHA8, GL_LUMINANCE16_ALPHA16, 0},
		{GL_INTENSITY4, GL_INTENSITY8, GL_INTENSITY16, 0},
		{GL_RGB4, GL_RGB8, GL_RGB16, 0},
		{GL_RGBA4, GL_RGBA8, GL_RGBA16, 0},
	};
	const GLenum p_float_format_array[][4] = {
		{0, 0, GL_ALPHA16F_ARB, GL_ALPHA32F_ARB},
		{0, 0, GL_LUMINANCE16F_ARB, GL_LUMINANCE32F_ARB},
		{0, 0, GL_LUMINANCE_ALPHA16F_ARB, GL_LUMINANCE_ALPHA32F_ARB},
		{0, 0, GL_INTENSITY16F_ARB, GL_INTENSITY32F_ARB},
		{0, 0, GL_RGB16F_ARB, GL_RGB32F_ARB},
		{0, 0, GL_RGBA16F_ARB, GL_RGBA32F_ARB},
	};

	int n_col_index;
	switch(n_bit_num) {
	case 4:
		n_col_index = 0;
		break;
	case 8:
		n_col_index = 1;
		break;
	case 16:
		n_col_index = 2;
		break;
	case 32:
		n_col_index = 3;
		break;
	default:
		return 0;
	}

	int n_row_index;
	switch(n_color_format) {
	case GL_ALPHA:
		n_row_index = 0;
		break;
	case GL_LUMINANCE:
		n_row_index = 1;
		break;
	case GL_LUMINANCE_ALPHA:
		n_row_index = 2;
		break;
	case GL_INTENSITY:
		n_row_index = 3;
		break;
	case GL_RGB:
		n_row_index = 4;
		break;
	case GL_RGBA:
		n_row_index = 5;
		break;
	default:
		return 0;
	}

	return ((b_float)? p_float_format_array : p_format_array)[n_row_index][n_col_index];
}

/*
 *								=== ~CGLRenderBuffer ===
 */

/*
 *								=== CGLRenderBuffer_FBO ===
 */

/*
 *	CGLRenderBuffer_FBO(int n_width, int n_height,
 *		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
 *		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
 *		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
 *		- constructor; creates the render buffer (OpenGL must be initialized)
 *		- n_width and n_height are dimensions of the output buffer
 *		- n_color_buffer_num is number of draw buffers (0 / 1 / 2 / 4 on geforce, 8 on quadro,
 *		  up to 16 in future; 0 means no color output is generated - depth map rendering)
 *		- b_color_texture_target enables binding color texture(s) as render target(s)
 *		  otherwise color is rendered into internal buffer (glReadPixels() / glCopyTexImage*())
 *		- n_color_internal_format is OpenGL internal texture format (such as GL_RGB8, ...)
 *		- b_depth_buffer enables depth output
 *		- b_depth_texture_target enables writing depth fragments into bound depth texture
 *		  (otherwise depth fragments are written into internal buffer)
 *		- n_depth_internal_format is depth buffer pixel format (one of GL_DEPTH_COMPONENT16,
 *		  GL_DEPTH_COMPONENT24 or GL_DEPTH_COMPONENT32)
 *		- b_stencil_buffer enables stencil output
 *		- b_stencil_texture_target enables writing stencil fragments into bound stencil texture
 *		  (otherwise stencil fragments are written into internal buffer)
 *		- n_stencil_internal_format is stencil buffer pixel format (one of
 *		  GL_STENCIL_INDEX1_EXT, GL_STENCIL_INDEX4_EXT, GL_STENCIL_INDEX8_EXT or
 *		  GL_STENCIL_INDEX16_EXT)
 */
CGLRenderBuffer_FBO::CGLRenderBuffer_FBO(int n_width, int n_height,
	int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
	bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
	bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
	:CGLRenderBuffer(n_width, n_height, n_color_buffer_num, b_color_texture_target,
	n_color_internal_format, b_depth_buffer, b_depth_texture_target, n_depth_internal_format,
	b_stencil_buffer, b_stencil_texture_target, n_stencil_internal_format)
{
	__FuncGuard("CGLRenderBuffer_FBO::CGLRenderBuffer_FBO");

	m_b_status = Create();
}

/*
 *	CGLRenderBuffer_FBO(int n_width, int n_height,
 *		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_format,
 *		int n_color_bit_num, bool b_float_color,
 *		bool b_depth_buffer, bool b_depth_texture_target, int n_depth_bit_num,
 *		bool b_stencil_buffer, bool b_stencil_texture_target, int n_stencil_bit_num)
 *		- constructor; creates the render buffer (OpenGL must be initialized)
 *		- n_width and n_height are dimensions of the output buffer
 *		- n_color_buffer_num is number of draw buffers (0 / 1 / 2 / 4 on geforce, 8 on quadro,
 *		  up to 16 in future; 0 means no color output is generated - depth map rendering)
 *		- b_color_texture_target enables binding color texture(s) as render target(s)
 *		  otherwise color is rendered into internal buffer (glReadPixels() / glCopyTexImage*())
 *		- n_color_format is OpenGL texture format (one of GL_RGB, GL_RGBA, GL_INTENSITY,
 *		  GL_LUMINANCE, GL_ALPHA, GL_LUMINANCE_ALPHA)
 *		- n_color_bit_num is color bit depth (4, 8 or 16 for integer, 16 or 32 for float formats)
 *		- b_float_color decides between integer (false) or float (true) formats
 *		- b_depth_buffer enables depth output
 *		- b_depth_texture_target enables writing depth fragments into bound depth texture
 *		  (otherwise depth fragments are written into internal buffer)
 *		- n_depth_bit_num is depth buffer pixel width (16, 24 or 32)
 *		- b_stencil_buffer enables stencil output
 *		- b_stencil_texture_target enables writing stencil fragments into bound stencil texture
 *		  (otherwise stencil fragments are written into internal buffer)
 *		- n_stencil_bit_num is stencil buffer pixel width (1, 4, 8 or 16)
 */
CGLRenderBuffer_FBO::CGLRenderBuffer_FBO(int n_width, int n_height,
	int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_format,
	int n_color_bit_num, bool b_float_color,
	bool b_depth_buffer, bool b_depth_texture_target, int n_depth_bit_num,
	bool b_stencil_buffer, bool b_stencil_texture_target, int n_stencil_bit_num)
	:CGLRenderBuffer(n_width, n_height, n_color_buffer_num, b_color_texture_target,
	n_color_format, n_color_bit_num, b_float_color, b_depth_buffer,
	b_depth_texture_target, n_depth_bit_num, b_stencil_buffer, b_stencil_texture_target,
	n_stencil_bit_num)
{
	__FuncGuard("CGLRenderBuffer_FBO::CGLRenderBuffer_FBO");

	if((n_color_buffer_num && !m_n_color_internal_format) ||
	   (b_depth_buffer && !m_n_depth_internal_format) ||
	   (b_stencil_buffer && !m_n_stencil_internal_format))
	    m_b_status = false;
	else
		m_b_status = Create();
}

/*
 *	virtual CGLRenderBuffer_FBO::~CGLRenderBuffer_FBO()
 *		- default destructor
 *		- takes care of cleaning up
 */
CGLRenderBuffer_FBO::~CGLRenderBuffer_FBO()
{
	__FuncGuard("CGLRenderBuffer_FBO::~CGLRenderBuffer_FBO");

	if(m_b_status) {
		if(m_n_draw_buffer_num && !m_b_color_texture_target)
			glDeleteRenderbuffersEXT(m_n_draw_buffer_num, m_p_color_rb);
		if(m_b_depth_buffer && !m_b_depth_texture_target)
			glDeleteRenderbuffersEXT(1, &m_n_depth_rb);
		if(m_b_stencil_buffer && !m_b_stencil_texture_target)
			glDeleteRenderbuffersEXT(1, &m_n_stencil_rb);
		glDeleteFramebuffersEXT(1, &m_n_framebuffer);

		m_b_status = false;
	}
}

/*
 *	static bool CGLRenderBuffer_FBO::b_Supported()
 *		- returns true in case the render buffer class have necessary hardware support
 */
bool CGLRenderBuffer_FBO::b_Supported()
{
	__FuncGuard("CGLRenderBuffer_FBO::b_Supported");

	return CGLExtensionHandler::b_SupportedExtension("GL_EXT_framebuffer_object") &&
		!CGLExtensionHandler::n_GetFramebufferObjectEXTFuncPointers();
}

/*
 *	static int CGLRenderBuffer_FBO::n_Max_DrawBuffer_Num()
 *		- returns maximal number of OpenGL draw buffers (1 where not supported, otherwise
 *		  1, 2, 4, up to 16 in future)
 */
int CGLRenderBuffer_FBO::n_Max_DrawBuffer_Num()
{
	__FuncGuard("CGLRenderBuffer_FBO::n_Max_DrawBuffer_Num");

	if(CGLExtensionHandler::b_Support_OpenGL(2, 0) &&
	   !CGLExtensionHandler::n_GetGL20FuncPointers()) {
		int n_value = 0;
		glGetIntegerv(GL_MAX_DRAW_BUFFERS, &n_value);
		int n_value2 = 0;
		glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &n_value2);
		if(glGetError() != GL_NO_ERROR)
			return 1;
		return (n_value > n_value2)? n_value2 : n_value;
	}
	return 1;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::HandleModeSwitch()
 *		- see if the render-buffer was lost during display mode switch
 *		- returns true in case render-buffer is ok or false in case render buffer
 *		  was lost (creates a new buffer in latter case, the result can be obtained
 *		  by call to b_Status())
 */
bool CGLRenderBuffer_FBO::HandleModeSwitch()
{
	__FuncGuard("CGLRenderBuffer_FBO::HandleModeSwitch");

	return true;
}

/*
 *	bool CGLRenderBuffer_FBO::Create()
 *		- create new FBO, according to params, passed to constructor
 *		- return true in case of success, otherwise false
 */
bool CGLRenderBuffer_FBO::Create()
{
	__FuncGuard("CGLRenderBuffer_FBO::Create");

	if(!CGLRenderBuffer_FBO::b_Supported() ||
	   CGLRenderBuffer_FBO::n_Max_DrawBuffer_Num() < m_n_draw_buffer_num)
		return false;
	_ASSERTE(m_n_draw_buffer_num <= 16);
	// check for support (get function pointers same time)

	glGenFramebuffersEXT(1, &m_n_framebuffer);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_n_framebuffer);
	if(glGetError() != GL_NO_ERROR)
		return false;
	// create a new framebuffer object

	if(m_n_draw_buffer_num && !m_b_color_texture_target) {
		glGenRenderbuffersEXT(m_n_draw_buffer_num, m_p_color_rb);
		for(int i = 0; i < m_n_draw_buffer_num; ++ i) {
			glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_p_color_rb[i]);
			glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, m_n_color_internal_format,
				m_n_width, m_n_height);
			glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
				GL_COLOR_ATTACHMENT0_EXT + i, GL_RENDERBUFFER_EXT, m_p_color_rb[i]);
		}
		if(glGetError() != GL_NO_ERROR)
			return false;
	}
	// create color buffer(s) (in case texture is not going to be bound only)

	if(m_b_depth_buffer && !m_b_depth_texture_target) {
		glGenRenderbuffersEXT(1, &m_n_depth_rb);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_n_depth_rb);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, m_n_depth_internal_format,
			m_n_width, m_n_height);
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
			GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, m_n_depth_rb);
		if(glGetError() != GL_NO_ERROR)
			return false;
	}
	// create depth buffer (in case depth texture is not going to be bound only)

	if(m_b_stencil_buffer && !m_b_stencil_texture_target) {
		glGenRenderbuffersEXT(1, &m_n_stencil_rb);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, m_n_stencil_internal_format,
			m_n_width, m_n_height);
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
			GL_STENCIL_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, m_n_stencil_rb);
		if(glGetError() != GL_NO_ERROR)
			return false;
	}
	// create stencil buffer (in case stencil texture is not going to be bound only)

	if(!(m_p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext())))
		return false;
	// create state-tracking object

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	m_b_active = false;
	// it was necessary to bind framebuffer at the beginning, unbind it now

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::Bind()
 *		- binds the render buffer as OpenGL output buffer
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::Bind()
{
	__FuncGuard("CGLRenderBuffer_FBO::Bind");

	if(!m_b_status)
		return false;

	if(m_b_active)
		return true;

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_n_framebuffer);
	if(m_n_draw_buffer_num != 1) {
		int n_max_draw_buffer_num = 0;
		glGetIntegerv(GL_MAX_DRAW_BUFFERS, &n_max_draw_buffer_num);
		_ASSERTE(n_max_draw_buffer_num <= 16);
		for(int i = 0; i < n_max_draw_buffer_num; ++ i)
			glGetIntegerv(GL_DRAW_BUFFER0_ARB + i, (int*)&m_p_draw_buffers[i]);
		// remember draw buffers state

		if(m_n_draw_buffer_num) {
			const GLenum p_draw_buffers[] = {
				GL_AUX0, GL_AUX0 + 1, GL_AUX0 + 2, GL_AUX0 + 3,
				GL_AUX0 + 4, GL_AUX0 + 5, GL_AUX0 + 6, GL_AUX0 + 7,
				GL_AUX0 + 8, GL_AUX0 + 9, GL_AUX0 + 10, GL_AUX0 + 11,
				GL_AUX0 + 12, GL_AUX0 + 13, GL_AUX0 + 14, GL_AUX0 + 15,
			};
			glDrawBuffers(m_n_draw_buffer_num, p_draw_buffers);
			// set new draw buffers state
		} else
			glDrawBuffer(GL_NONE);
	}

	m_b_active = true;

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::Release()
 *		- releases the render buffer from OpenGL output buffer binding
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::Release()
{
	__FuncGuard("CGLRenderBuffer_FBO::Release");

	if(!m_b_status || !m_b_active)
		return false;

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	if(m_n_draw_buffer_num > 1) {
		int n_max_draw_buffer_num = 0;
		glGetIntegerv(GL_MAX_DRAW_BUFFERS, &n_max_draw_buffer_num);
		_ASSERTE(n_max_draw_buffer_num <= 16);
		glDrawBuffers(n_max_draw_buffer_num, m_p_draw_buffers);
		// restore draw buffers state
	}

	m_b_active = false;

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_2D(const CGLTexture &r_texture,
 *		GLenum n_texture_type, int n_index = 0)
 *		- binds OpenGL texture r_texture.n_Id() as target for color output
 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
 *		- n_index is index of one of draw buffers texture is to be bound to
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindColorTexture_2D(const CGLTexture &r_texture,
	GLenum n_texture_type, int n_index)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindColorTexture_2D");

	if(!m_b_status || !m_b_active || !m_n_draw_buffer_num)
		return false;
	if(n_index < 0 || n_index >= m_n_draw_buffer_num)
		return false;

	if(m_p_bound_color_texture[n_index] == r_texture.n_Id())
		return true;
	// bound already

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
		GL_COLOR_ATTACHMENT0_EXT + n_index, n_texture_type, r_texture.n_Id(), 0);

	if(glGetError() != GL_NO_ERROR)
		return false;

	if(!m_p_bound_color_texture[n_index] && r_texture.n_Id())
		m_n_bound_color_texture_num ++;
	else if(m_p_bound_color_texture[n_index] && !r_texture.n_Id())
		m_n_bound_color_texture_num --;
	// maintain number of bound textures

	m_p_bound_color_texture[n_index] = r_texture.n_Id();

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_3D(const CGLTexture &r_texture,
 *		int n_z_offset, Glenum int n_index = 0)
 *		- binds OpenGL texture r_texture.n_Id() as target for color output
 *		- n_z_offset is index of 2D slice to be rendered to
 *		- n_index is index of one of draw buffers texture is to be bound to
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindColorTexture_3D(const CGLTexture &r_texture, int n_z_offset, int n_index)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindColorTexture_3D");

	if(!m_b_status || !m_b_active || !m_n_draw_buffer_num)
		return false;
	if(n_index < 0 || n_index >= m_n_draw_buffer_num)
		return false;

	if(m_p_bound_color_texture[n_index] == r_texture.n_Id())
		return true;
	// bound already

	glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,
		GL_COLOR_ATTACHMENT0_EXT + n_index, GL_TEXTURE_3D, r_texture.n_Id(), 0, n_z_offset);

	if(glGetError() != GL_NO_ERROR)
		return false;

	if(!m_p_bound_color_texture[n_index] && r_texture.n_Id())
		m_n_bound_color_texture_num ++;
	else if(m_p_bound_color_texture[n_index] && !r_texture.n_Id())
		m_n_bound_color_texture_num --;
	// maintain number of bound textures

	m_p_bound_color_texture[n_index] = r_texture.n_Id();

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindDepthTexture_2D(const CGLTexture &r_texture,
 *		GLenum n_texture_type)
 *		- binds OpenGL texture r_texture.n_Id() as target for depth output
 *		- n_texture_type is either GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindDepthTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindDepthTexture_2D");

	if(!m_b_status || !m_b_active || !m_b_depth_buffer || !m_b_depth_texture_target)
		return false;

	if(m_n_bound_depth_texture == r_texture.n_Id())
		return true;
	// bound already

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
		GL_DEPTH_ATTACHMENT_EXT, n_texture_type, r_texture.n_Id(), 0);

	if(glGetError() != GL_NO_ERROR)
		return false;

	m_n_bound_depth_texture = r_texture.n_Id();

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindStencilTexture_2D(const CGLTexture &r_texture,
 *		GLenum n_texture_type)
 *		- binds OpenGL texture r_texture.n_Id() as target for stencil output
 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindStencilTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindStencilTexture_2D");

	if(!m_b_status || !m_b_active || !m_b_stencil_buffer || !m_b_stencil_texture_target)
		return false;

	if(m_n_bound_stencil_texture == r_texture.n_Id())
		return true;
	// bound already

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
		GL_STENCIL_ATTACHMENT_EXT, n_texture_type, r_texture.n_Id(), 0);

	if(glGetError() != GL_NO_ERROR)
		return false;

	m_n_bound_stencil_texture = r_texture.n_Id();

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_2D(GLuint n_texture,
 *		GLenum n_texture_type, int n_index = 0)
 *		- binds OpenGL texture n_texture as target for color output
 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
 *		- n_index is index of one of draw buffers texture is to be bound to
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindColorTexture_2D(GLuint n_texture,
	GLenum n_texture_type, int n_index)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindColorTexture_2D");

	if(!m_b_status || !m_b_active || !m_n_draw_buffer_num)
		return false;
	if(n_index < 0 || n_index >= m_n_draw_buffer_num)
		return false;

	if(m_p_bound_color_texture[n_index] == n_texture)
		return true;
	// bound already

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
		GL_COLOR_ATTACHMENT0_EXT + n_index, n_texture_type, n_texture, 0);

	if(glGetError() != GL_NO_ERROR)
		return false;

	if(!m_p_bound_color_texture[n_index] && n_texture)
		m_n_bound_color_texture_num ++;
	else if(m_p_bound_color_texture[n_index] && !n_texture)
		m_n_bound_color_texture_num --;
	// maintain number of bound textures

	m_p_bound_color_texture[n_index] = n_texture;

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_3D(GLuint n_texture,
 *		int n_z_offset, Glenum int n_index = 0)
 *		- binds OpenGL texture n_texture as target for color output
 *		- n_z_offset is index of 2D slice to be rendered to
 *		- n_index is index of one of draw buffers texture is to be bound to
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindColorTexture_3D(GLuint n_texture, int n_z_offset, int n_index)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindColorTexture_3D");

	if(!m_b_status || !m_b_active || !m_n_draw_buffer_num)
		return false;
	if(n_index < 0 || n_index >= m_n_draw_buffer_num)
		return false;

	if(m_p_bound_color_texture[n_index] == n_texture)
		return true;
	// bound already

	glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,
		GL_COLOR_ATTACHMENT0_EXT + n_index, GL_TEXTURE_3D, n_texture, 0, n_z_offset);

	if(glGetError() != GL_NO_ERROR)
		return false;

	if(!m_p_bound_color_texture[n_index] && n_texture)
		m_n_bound_color_texture_num ++;
	else if(m_p_bound_color_texture[n_index] && !n_texture)
		m_n_bound_color_texture_num --;
	// maintain number of bound textures

	m_p_bound_color_texture[n_index] = n_texture;

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindDepthTexture_2D(GLuint n_texture,
 *		GLenum n_texture_type)
 *		- binds OpenGL texture n_texture as target for depth output
 *		- n_texture_type is either GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindDepthTexture_2D(GLuint n_texture, GLenum n_texture_type)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindDepthTexture_2D");

	if(!m_b_status || !m_b_active || !m_b_depth_buffer || !m_b_depth_texture_target)
		return false;

	if(m_n_bound_depth_texture == n_texture)
		return true;
	// bound already

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
		GL_DEPTH_ATTACHMENT_EXT, n_texture_type, n_texture, 0);

	if(glGetError() != GL_NO_ERROR)
		return false;

	m_n_bound_depth_texture = n_texture;

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::BindStencilTexture_2D(GLuint n_texture,
 *		GLenum n_texture_type)
 *		- binds OpenGL texture n_texture as target for stencil output
 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
 *		- binding texture with id 0 releases previously bound texture
 *		- returns true on success or false on failure
 */
bool CGLRenderBuffer_FBO::BindStencilTexture_2D(GLuint n_texture, GLenum n_texture_type)
{
	__FuncGuard("CGLRenderBuffer_FBO::BindStencilTexture_2D");

	if(!m_b_status || !m_b_active || !m_b_stencil_buffer || !m_b_stencil_texture_target)
		return false;

	if(m_n_bound_stencil_texture == n_texture)
		return true;
	// bound already

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
		GL_STENCIL_ATTACHMENT_EXT, n_texture_type, n_texture, 0);

	if(glGetError() != GL_NO_ERROR)
		return false;

	m_n_bound_stencil_texture = n_texture;

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::b_DirectRenderToTexture() const
 *		- returns true in case direct rendering to all textures, being bound is supported,
 *		  otherwise false
 *		- note in this implementation it always returns true
 */
bool CGLRenderBuffer_FBO::b_DirectRenderToTexture() const
{
	__FuncGuard("CGLRenderBuffer_FBO::b_DirectRenderToTexture");

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::CopyTextureData(bool b_release_textures)
 *		- copies texture data to textures it wasn't possible to directly render to
 *		- in case b_release_textures is true, all currently bound textures are released
 *		- returns true on success or false on failure
 *		- note in this implementation it does not copy no data and always returns true
 */
bool CGLRenderBuffer_FBO::CopyTextureData(bool b_release_textures)
{
	__FuncGuard("CGLRenderBuffer_FBO::CopyTextureData");

	if(!m_b_status || !m_b_active)
		return false;

	if(b_release_textures) {
		for(int i = 0; i < m_n_draw_buffer_num && m_n_bound_color_texture_num > 0; ++ i) {
			if(m_p_bound_color_texture[i]) {
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
					GL_COLOR_ATTACHMENT0_EXT + i, GL_TEXTURE_2D, 0, 0);
				m_p_bound_color_texture[i] = 0;
				m_n_bound_color_texture_num --;
			}
		}
		_ASSERTE(!m_n_bound_color_texture_num);
		if(m_b_depth_buffer && m_b_depth_texture_target && m_n_bound_depth_texture) {
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
				GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, 0, 0);
			m_n_bound_depth_texture = 0;
		}
		if(m_b_stencil_buffer && m_b_stencil_texture_target && m_n_bound_stencil_texture) {
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
				GL_STENCIL_ATTACHMENT_EXT, GL_TEXTURE_2D, 0, 0);
			m_n_bound_stencil_texture = 0;
		}
	}

	return true;
}

/*
 *	virtual bool CGLRenderBuffer_FBO::b_Status() const
 *		- returns true in case render buffer, specified by constructor parameters
 *		  was successfuly created and is ready to use
 */
bool CGLRenderBuffer_FBO::b_Status() const
{
	__FuncGuard("CGLRenderBuffer_FBO::b_Status");

	return m_b_status && (!m_b_active ||
		glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) == GL_FRAMEBUFFER_COMPLETE_EXT);
}

/*
 *	virtual CGLState *CGLRenderBuffer_FBO::p_State()
 *		- returns render-buffer OpenGL state guard (should be identic to
 *		  global state guard for FBO, different for PBO)
 *		- returns 0 in case render buffer is not initialized / ready
 */
CGLState *CGLRenderBuffer_FBO::p_State()
{
	__FuncGuard("CGLRenderBuffer_FBO::p_State");

	return m_p_state;
}

/*
 *								=== ~CGLRenderBuffer_FBO ===
 */
